import doctest
import pytest

from insights.core.exceptions import ParseException, SkipComponent
from insights.parsers import grubby
from insights.parsers.grubby import GrubbyDefaultIndex, GrubbyDefaultKernel, GrubbyInfoAll
from insights.tests import context_wrap

DEFAULT_INDEX_1 = '0'
DEFAULT_INDEX_2 = '1'
ABDEFAULT_INDEX_EMPTY = ''
DEFAULT_INDEX_AB = '-2'
DEFAULT_INDEX_W_ERR_MSG = """
/usr/libexec/grubby/grubby-bls: line 37: /etc/machine-id: No such file or directory
0
""".strip()
DEFAULT_INDEX_AB_W_ERR_MSG = """
/usr/libexec/grubby/grubby-bls: line 37: /etc/machine-id: No such file or directory
some-other-not-index-line
""".strip()

DEFAULT_KERNEL = "/boot/vmlinuz-2.6.32-573.el6.x86_64"
DEFAULT_KERNEL_EMPTY = ""
DEFAULT_KERNEL_AB = """
/boot/vmlinuz-2.6.32-573.el6.x86_64"
/boot/vmlinuz-2.6.32-573.el6.x86_64"
""".strip()
DEFAULT_KERNEL_INVALID = 'rpm-sort: Invalid input'
DEFAULT_KERNEL_INVALID_2 = 'rpm-sort: Invalid input '
DEFAULT_KERNEL_INVALID_3 = """
/usr/libexec/grubby/grubby-bls: fork: retry: Resource temporarily unavailable
/usr/libexec/grubby/grubby-bls: fork: retry: Resource temporarily unavailable
""".strip()

DEFAULT_KERNEL_WITH_ERRORS_MSGS_1 = """
/usr/libexec/grubby/grubby-bls: fork: retry: Resource temporarily unavailable
/usr/libexec/grubby/grubby-bls: fork: retry: Resource temporarily unavailable
/usr/libexec/grubby/grubby-bls: fork: retry: Resource temporarily unavailable
/usr/libexec/grubby/grubby-bls: fork: retry: Resource temporarily unavailable
/boot/vmlinuz-4.18.0-147.5.1.el8_1.x86_64
""".strip()
DEFAULT_KERNEL_WITH_ERRORS_MSGS_2 = """
grub2-editenv: error: cannot rename the file /boot/grub2/grubenv.new to /boot/grub2/grubenv: No such file or directory.
/boot/vmlinuz-3.10.0-862.el7.x86_64
""".strip()
DEFAULT_KERNEL_WITH_ERRORS_MSGS_3 = """
grep: /boot/grub2/grubenv: No such file or directory
/boot/cm/images/dell-rhel8.2/boot/vmlinuz-4.18.0-348.23.1.el8_5.x86_64
""".strip()
DEFAULT_KERNEL_WITH_ERRORS_MSGS_4 = """
/etc/os-release: line 5: VERSION_ID-Peter=8.7: command not found
/etc/os-release: line 6: PLATFORM_ID-Peter=platform:el8: command not found
/boot/vmlinuz-4.18.0-425.10.1.el8_7.x86_64
""".strip()
DEFAULT_KERNEL_WITH_ERRORS_MSGS_5 = """
/usr/libexec/grubby/grubby-bls: line 37: /etc/machine-id: No such file or directory
/boot/vmlinuz-4.18.0-553.56.1.el8_10.x86_64
""".strip()

GRUBBY_INFO_ALL_1 = """
index=0
kernel="/boot/vmlinuz-5.14.0-162.6.1.el9_1.x86_64"
args="ro crashkernel=1G-4G:192M,4G-64G:256M,64G-:512M resume=/dev/mapper/rhel-swap rd.lvm.lv=rhel/root rd.lvm.lv=rhel/swap rhgb quiet retbleed=stuff"
root="/dev/mapper/rhel-root"
initrd="/boot/initramfs-5.14.0-162.6.1.el9_1.x86_64.img"
title="Red Hat Enterprise Linux (5.14.0-162.6.1.el9_1.x86_64) 9.1 (Plow)"
id="4d684a4a6166439a867e701ded4f7e10-5.14.0-162.6.1.el9_1.x86_64"
index=1
kernel="/boot/vmlinuz-5.14.0-70.13.1.el9_0.x86_64"
args="ro crashkernel=1G-4G:192M,4G-64G:256M,64G-:512M resume=/dev/mapper/rhel-swap rd.lvm.lv=rhel/root rd.lvm.lv=rhel/swap rhgb quiet retbleed=stuff"
root="/dev/mapper/rhel-root"
initrd="/boot/initramfs-5.14.0-70.13.1.el9_0.x86_64.img"
title="Red Hat Enterprise Linux (5.14.0-70.13.1.el9_0.x86_64) 9.0 (Plow)"
id="4d684a4a6166439a867e701ded4f7e10-5.14.0-70.13.1.el9_0.x86_64"
""".strip()

GRUBBY_INFO_ALL_2 = """
index=0
kernel=/boot/vmlinuz-3.10.0-862.el7.x86_64
args="ro console=tty0 console=ttyS0,115200n8 no_timer_check net.ifnames=0 crashkernel=auto LANG=en_US.UTF-8"
root=UUID=6bea2b7b-e6cc-4dba-ac79-be6530d348f5
initrd=/boot/initramfs-3.10.0-862.el7.x86_64.img
title=Red Hat Enterprise Linux Server (3.10.0-862.el7.x86_64) 7.5 (Maipo)
index=1
kernel=/boot/vmlinuz-0-rescue-1b461b2e96854984bc0777c4b4b518a9
args="ro console=tty0 console=ttyS0,115200n8 no_timer_check net.ifnames=0 crashkernel=auto"
root=UUID=6bea2b7b-e6cc-4dba-ac79-be6530d348f5
initrd=/boot/initramfs-0-rescue-1b461b2e96854984bc0777c4b4b518a9.img
title=Red Hat Enterprise Linux Server (0-rescue-1b461b2e96854984bc0777c4b4b518a9) 7.5 (Maipo)
index=2
non linux entry
""".strip()

GRUBBY_INFO_ALL_INVALID_1 = """
some head lines
index=0
non linux entry

some tail lines
""".strip()
GRUBBY_INFO_ALL_INVALID_2 = """
some head lines
kernel=/boot/vmlinuz-3.10.0-862.el7.x86_64
args="ro console=tty0 console=ttyS0,115200n8 no_timer_check net.ifnames=0 crashkernel=auto LANG=en_US.UTF-8"
root=UUID=6bea2b7b-e6cc-4dba-ac79-be6530d348f5
initrd=/boot/initramfs-3.10.0-862.el7.x86_64.img
title=Red Hat Enterprise Linux Server (3.10.0-862.el7.x86_64) 7.5 (Maipo)
some tail lines
""".strip()
GRUBBY_INFO_ALL_INVALID_2 = """
some head lines
index=some-index
kernel=/boot/vmlinuz-3.10.0-862.el7.x86_64
args="ro console=tty0 console=ttyS0,115200n8 no_timer_check net.ifnames=0 crashkernel=auto LANG=en_US.UTF-8"
root=UUID=6bea2b7b-e6cc-4dba-ac79-be6530d348f5
initrd=/boot/initramfs-3.10.0-862.el7.x86_64.img
title=Red Hat Enterprise Linux Server (3.10.0-862.el7.x86_64) 7.5 (Maipo)
some tail lines
""".strip()


def test_grubby_default_index():
    res = GrubbyDefaultIndex(context_wrap(DEFAULT_INDEX_1))
    assert res.default_index == 0

    res = GrubbyDefaultIndex(context_wrap(DEFAULT_INDEX_2))
    assert res.default_index == 1

    res = GrubbyDefaultIndex(context_wrap(DEFAULT_INDEX_W_ERR_MSG))
    assert res.default_index == 0


def test_grubby_default_index_ab():
    with pytest.raises(SkipComponent) as excinfo:
        GrubbyDefaultIndex(context_wrap(ABDEFAULT_INDEX_EMPTY))
    assert 'Empty output' in str(excinfo.value)

    with pytest.raises(ParseException) as excinfo:
        GrubbyDefaultIndex(context_wrap(DEFAULT_INDEX_AB))
    assert 'Invalid output:' in str(excinfo.value)

    with pytest.raises(ParseException) as excinfo:
        GrubbyDefaultIndex(context_wrap(DEFAULT_INDEX_AB_W_ERR_MSG))
    assert 'Invalid output:' in str(excinfo.value)


def test_grubby_default_kernel():
    res = GrubbyDefaultKernel(context_wrap(DEFAULT_KERNEL))
    assert res.default_kernel == DEFAULT_KERNEL

    content_with_error_msgs = [
        DEFAULT_KERNEL_WITH_ERRORS_MSGS_1,
        DEFAULT_KERNEL_WITH_ERRORS_MSGS_2,
        DEFAULT_KERNEL_WITH_ERRORS_MSGS_3,
        DEFAULT_KERNEL_WITH_ERRORS_MSGS_4,
        DEFAULT_KERNEL_WITH_ERRORS_MSGS_5,
    ]
    for content in content_with_error_msgs:
        this_res = GrubbyDefaultKernel(context_wrap(content))
        assert this_res.default_kernel == content.split()[-1].strip()


def test_grubby_default_kernel_ab():
    with pytest.raises(SkipComponent) as excinfo:
        GrubbyDefaultKernel(context_wrap(DEFAULT_KERNEL_EMPTY))
    assert 'Empty output' in str(excinfo.value)

    with pytest.raises(ParseException) as excinfo:
        GrubbyDefaultKernel(context_wrap(DEFAULT_KERNEL_AB))
    assert 'Invalid output:' in str(excinfo.value)

    with pytest.raises(ParseException) as excinfo:
        GrubbyDefaultKernel(context_wrap(DEFAULT_KERNEL_INVALID))
    assert 'Invalid output:' in str(excinfo.value)

    with pytest.raises(ParseException) as excinfo:
        GrubbyDefaultKernel(context_wrap(DEFAULT_KERNEL_INVALID_2))
    assert 'Invalid output:' in str(excinfo.value)

    with pytest.raises(ParseException) as excinfo:
        GrubbyDefaultKernel(context_wrap(DEFAULT_KERNEL_INVALID_3))
    assert 'Invalid output:' in str(excinfo.value)


def test_grubby_info_all():
    res = GrubbyInfoAll(context_wrap(GRUBBY_INFO_ALL_1))

    assert res.unparsed_lines == []
    assert len(res.boot_entries) == 2
    assert res.boot_entries[0] == dict(
        index=0,
        kernel="/boot/vmlinuz-5.14.0-162.6.1.el9_1.x86_64",
        args={
            'ro': [True],
            'crashkernel': ['1G-4G:192M,4G-64G:256M,64G-:512M'],
            'resume': ['/dev/mapper/rhel-swap'],
            'rd.lvm.lv': ['rhel/root', 'rhel/swap'],
            'rhgb': [True],
            'quiet': [True],
            'retbleed': ['stuff'],
        },
        raw_args="ro crashkernel=1G-4G:192M,4G-64G:256M,64G-:512M resume=/dev/mapper/rhel-swap rd.lvm.lv=rhel/root rd.lvm.lv=rhel/swap rhgb quiet retbleed=stuff",
        root="/dev/mapper/rhel-root",
        initrd="/boot/initramfs-5.14.0-162.6.1.el9_1.x86_64.img",
        title="Red Hat Enterprise Linux (5.14.0-162.6.1.el9_1.x86_64) 9.1 (Plow)",
        id="4d684a4a6166439a867e701ded4f7e10-5.14.0-162.6.1.el9_1.x86_64",
    )

    assert "kernel" in res.boot_entries[1]
    assert res.boot_entries[1]["kernel"] == "/boot/vmlinuz-5.14.0-70.13.1.el9_0.x86_64"
    assert res.boot_entries[1].get("root") == "/dev/mapper/rhel-root"

    entry_args = res.boot_entries[1].get("args")
    assert entry_args.get("ro") == [True]
    assert entry_args.get("rd.lvm.lv") == ['rhel/root', 'rhel/swap']

    res = GrubbyInfoAll(context_wrap(GRUBBY_INFO_ALL_2))
    assert len(res.boot_entries) == 2
    assert res.unparsed_lines == ["non linux entry"]


def test_grubby_info_all_ab():
    with pytest.raises(SkipComponent) as excinfo:
        GrubbyInfoAll(context_wrap(""))
    assert 'Empty output' in str(excinfo.value)

    with pytest.raises(SkipComponent) as excinfo:
        GrubbyInfoAll(context_wrap(GRUBBY_INFO_ALL_INVALID_1))
    assert 'No valid entry parsed' in str(excinfo.value)

    with pytest.raises(ParseException) as excinfo:
        GrubbyInfoAll(context_wrap(GRUBBY_INFO_ALL_INVALID_2))
    assert 'Invalid index value:' in str(excinfo.value)


def test_doc_examples():
    env = {
        'grubby_default_index': GrubbyDefaultIndex(context_wrap(DEFAULT_INDEX_1)),
        'grubby_default_kernel': GrubbyDefaultKernel(context_wrap(DEFAULT_KERNEL)),
        'grubby_info_all': GrubbyInfoAll(context_wrap(GRUBBY_INFO_ALL_1)),
    }
    failed, total = doctest.testmod(grubby, globs=env)
    assert failed == 0



"""
ContainersInspect - commands ``docker|podman inspect``
======================================================

This parser reads the output of commands: "/usr/bin/docker|podman inspect <containers ID>"
which are used to show the metadata information of containers.
"""
from insights import parser, JSONParser
from insights.specs import Specs


@parser(Specs.containers_inspect)
class ContainersInspect(JSONParser):
    """
    Class for parsing the output of the containers inspect commands
    ``/usr/bin/docker|podman inspect <containers ID>``


    Typical Output of this command after datasource containers_inspect is::

        [
            {
                "Id": "aeaea3ead527",
                "Image": "538460c14d75dee1504e56ad8ddb7fe039093b1530ef8f90442a454b9aa3dc8b",
                "engine": "podman",
                "HostConfig|Privileged": false,
                "Config|Cmd": ["sleep", "1000000"]
            }
        ]

    Attributes:
        data (list): A list containing the parsed information

    Examples:
        >>> from insights.core.filters import add_filter
        >>> from insights.specs import Specs
        >>> add_filter(Specs.container_inspect_keys, ['HostConfig|Privileged'])
        >>> str(inspect_containers.data[0]["Id"])
        'aeaea3ead527'
        >>> str(inspect_containers.data[0]["engine"])
        'podman'
        >>> inspect_containers.data[0]["HostConfig|Privileged"]
        False
    """

    def parse_content(self, content):
        super(ContainersInspect, self).parse_content(content[0])



"""
NVIDIA Related Information
==========================

NvidiaSmiL - command ``/usr/bin/nvidia-smi -L``
-----------------------------------------------

NvidiaSmiActiveClocksEventReasons - command ``/usr/bin/nvidia-smi --query-gpu=name,clocks_event_reasons.active --format=csv,noheader``
--------------------------------------------------------------------------------------------------------------------------------------

NvidiaSmiQueryGPU - command ``/usr/bin/nvidia-smi --query-gpu=index,name,uuid,memory.total --format=csv,noheader``
------------------------------------------------------------------------------------------------------------------
"""

from collections import namedtuple
from insights.core import Parser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs
from insights.util import deprecated


NvidiaGPUInfo = namedtuple(
    "NvidiaGPUInfo",
    ['index', 'model', 'uuid', 'memory_total'],
)
"""namedtuple: Represents the information parsed from ``nvidia-smi --query-gpu`` command output."""


# Refer to the following doc for the detailed bitmask of active clock event reasons:
# - https://docs.nvidia.com/deploy/nvml-api/group__nvmlClocksEventReasons.html
BITMASK = {
    "none": 0x0000000000000000,
    "gpu_idle": 0x0000000000000001,
    "applications_clocks_setting": 0x0000000000000002,
    "sw_power_cap": 0x0000000000000004,
    "hw_slowdown": 0x0000000000000008,
    "sync_boost": 0x0000000000000010,
    "sw_thermal_slowdown": 0x0000000000000020,
    "hw_thermal_slowdown": 0x0000000000000040,
    "hw_power_brake_slowdown": 0x0000000000000080,
    "display_clock_setting": 0x0000000000000100,
}


@parser(Specs.nvidia_smi_l)
class NvidiaSmiL(Parser, list):
    """

    .. warning::
        This class is deprecated and will be removed from 3.8.0.
        Please use the :class:`insights.parsers.nvidia.NvidiaSmiQueryGPU` instead.

    Prase for output of command `/usr/bin/nvidia-smi -L`. This command lists
    each of the NVIDIA GPUs in the system, along with their UUIDs.

    The GPU info shown in each line will be parsed as follows::

        model (string):   The gpu model
        uuid (string):    The gpu uuid

    Raises:
        ParseException: When run into unparsable gpu line
        SkipComponent: When content is empty or no parsable content

    Sample Content::

        GPU 0: NVIDIA A100-PCIE-40GB (UUID: GPU-63110aaa-3561-c8f5-e125-4ab40bbcf838)
        GPU 1: NVIDIA A100-PCIE-40GB (UUID: GPU-c9bd25dc-c0c4-3ab6-8f7f-3ad16d6bde4a)

    Examples::
        >>> gpus.gpu_count
        2
        >>> "NVIDIA A100-PCIE-40GB" in gpus.gpu_models
        True
        >>> gpus[0]
        {'model': 'NVIDIA A100-PCIE-40GB', 'uuid': 'GPU-63110aaa-3561-c8f5-e125-4ab40bbcf838'}
    """

    def __init__(self, context):
        deprecated(
            NvidiaSmiL,
            "Please use the :class:`insights.parsers.nvidia.NvidiaSmiQueryGPU` instead.",
            "3.8.0",
        )
        super(NvidiaSmiL, self).__init__(context)

    def parse_content(self, content):
        if not content:
            raise SkipComponent("Empty content")

        _gpu_index_validator = 0

        for line in content:
            if line.startswith("GPU ") and "(UUID: " in line and line.endswith(')'):
                spl_line = line.split("(UUID: ")
                spl_left = spl_line[0].split(': ')
                if not len(spl_left) == 2:
                    raise ParseException("Unparsable GPU model: %s" % line)

                gpu_id = spl_left[0].split()[-1]
                if not (gpu_id.isdigit() and int(gpu_id) == _gpu_index_validator):
                    raise ParseException("Unparsable GPU id: %s" % line)
                _gpu_index_validator += 1

                gpu_model = spl_left[-1].strip()
                gpu_uuid = spl_line[-1].strip(') ')
                if not (gpu_model and gpu_uuid):
                    raise ParseException("Unparsable GPU line: %s" % line)

                self.append(
                    {
                        "model": gpu_model,
                        "uuid": gpu_uuid,
                    }
                )

        if len(self) < 1:
            raise ParseException("Empty GPU info after parse: %s" % content)

    @property
    def gpu_count(self):
        """
        str: Returns the GPU count
        """
        return len(self)

    @property
    def gpu_models(self):
        """
        str: Returns the GPUs model set
        """
        return set([gpu["model"] for gpu in self])


@parser(Specs.nvidia_smi_active_clocks_event_reasons)
class NvidiaSmiActiveClocksEventReasons(Parser, list):
    """
    Parser for the output of command `/usr/bin/nvidia-smi --query-gpu=name,clocks_event_reasons.active --format=csv,noheader`.
    This command lists each of the NVIDIA GPUs in the system, along with their names and bitmasks
    of active clock event reasons.

    Raises:
        ParseException: When run into an unparsable line
        SkipComponent: When content is empty

    Sample Content::
        NVIDIA L4, 0x0000000000000001
        NVIDIA A1, 0x0000000000000000
        NVIDIA H1, 0x0000000000000084

    Examples::
        >>> type(active_clocks_event_reasons)
        <class 'insights.parsers.nvidia.NvidiaSmiActiveClocksEventReasons'>
        >>> len(active_clocks_event_reasons)
        3
        >>> active_clocks_event_reasons[0]['applications_clocks_setting']
        False
        >>> active_clocks_event_reasons[2]['hw_power_brake_slowdown']
        True
        >>> active_clocks_event_reasons[2]['none']
        False
    """

    def parse_content(self, content):
        if not content:
            raise SkipComponent("Empty content.")

        for line in content:
            items = line.split(",")
            if len(items) != 2 or not items[1].strip().startswith("0x"):
                raise ParseException(
                    "Not an expected command output for active clocks event reasons: %s" % line
                )
            bitmask = int(items[1].strip().strip("LL"), 16)

            data = dict(gpu_name=items[0].strip())
            for key, bm in BITMASK.items():
                if key == "none":
                    data[key] = bm | bitmask == bm
                else:
                    data[key] = bm & bitmask == bm
            self.append(data)


@parser(Specs.nvidia_smi_query_gpu)
class NvidiaSmiQueryGPU(Parser, list):
    """
    Parser for the output of command
    `/usr/bin/nvidia-smi --query-gpu=index,name,uuid,memory.total --format=csv,noheader`.
    This command lists each of the NVIDIA GPUs info in the system.

    The GPU info shown in each line will be parsed to a namedtuple ``NvidiaGPUInfo``
    contains the follows info::

        index (str):        The gpu index
        model (str):        The gpu model
        uuid (str):         The gpu uuid
        memory_total(str):  The gpu total memory

    Raises:
        ParseException: When run into an unparsable line
        SkipComponent: When content is empty

    Sample Content::

        0, NVIDIA L4, GPU-24598a07-f97d-fc79-86de-485c4c82d01c, 23034 MiB
        1, NVIDIA A1, GPU-63110aaa-3561-c8f5-e125-4ab40bbcf838, 16280 MiB
        2, NVIDIA H1, GPU-c9bd25dc-c0c4-3ab6-8f7f-3ad16d6bde4a, 24564 MiB

    Examples::
        >>> gpus_info.gpu_count
        3
        >>> "NVIDIA L4" in gpus_info.gpu_models
        True
        >>> type(gpus_info[0])
        <class 'insights.parsers.nvidia.NvidiaGPUInfo'>
        >>> gpus_info[0].uuid
        'GPU-24598a07-f97d-fc79-86de-485c4c82d01c'
        >>> gpus_info[0].memory_total
        '23034 MiB'
    """

    _columns = NvidiaGPUInfo._fields

    def parse_content(self, content):
        for line in content:
            items = [v.strip() for v in line.strip().split(",")]
            if len(items) == len(self._columns) and all(items):
                self.append(NvidiaGPUInfo(*items))
            else:
                raise ParseException("Not an expected command output: %s" % line)

        if len(self) == 0:
            raise SkipComponent

    @property
    def gpu_count(self):
        """
        str: Returns the GPU count
        """
        return len(self)

    @property
    def gpu_models(self):
        """
        str: Returns the GPUs model set
        """
        return set([gpu.model for gpu in self])



# -*- coding: UTF-8 -*-

import six
import mock
import pytest
from .helpers import insights_upload_conf
from mock.mock import patch, Mock, call
from insights.client.collection_rules import correct_format, load_yaml, verify_permissions


conf_remove_file = '/tmp/remove.conf'
conf_file_redaction_file = '/tmp/file-redaction.yaml'
conf_file_content_redaction_file = '/tmp/file-content-redaction.yaml'
removed_files = ["/etc/some_file", "/tmp/another_file"]


def patch_open(filedata):
    if six.PY3:
        open_name = 'builtins.open'
    else:
        open_name = '__builtin__.open'

    return patch(open_name, mock.mock_open(read_data=filedata), create=True)


# Tests for the correct_format function
def test_correct_format_ok_validtypes():
    '''
    Verify that valid config is allowed when
    proper keys and lists of strings are specified
    '''
    # files and commands (file-redaction.yaml)
    parsed_data = {
        'commands': ['/bin/test', '/bin/test2'],
        'files': ['/var/lib/aaa', '/var/lib/nnn'],
    }
    expected_keys = ('commands', 'files')
    err, msg = correct_format(parsed_data, expected_keys, conf_file_redaction_file)
    assert not err
    assert msg is None

    # patterns w. list of strings (file-content-redaction.yaml)
    parsed_data = {'patterns': ['abcd', 'bcdef'], 'keywords': ['example', 'example2']}
    expected_keys = ('patterns', 'keywords')
    err, msg = correct_format(parsed_data, expected_keys, conf_file_content_redaction_file)
    assert not err
    assert msg is None

    # patterns w. regex object (file-content-redaction.yaml)
    parsed_data = {'patterns': {'regex': ['abcd', 'bcdef']}, 'keywords': ['example', 'example2']}
    expected_keys = ('patterns', 'keywords')
    err, msg = correct_format(parsed_data, expected_keys, conf_file_content_redaction_file)
    assert not err
    assert msg is None


def test_config_verification_ok_emptyvalues():
    '''
    Verify that valid config is allowed when
    proper keys and empty (None) values are specified
    '''
    parsed_data = {'patterns': None, 'keywords': ['abc', 'def']}
    expected_keys = ('patterns', 'keywords')
    err, msg = correct_format(parsed_data, expected_keys, conf_file_content_redaction_file)
    assert not err
    assert msg is None


def test_config_verification_bad_invalidkeys():
    '''
    Verify that a config with invalid keys is not allowed
    '''
    parsed_data = {'commands': None, 'files': None, 'somekey': None}
    expected_keys = ('commands', 'files')
    err, msg = correct_format(parsed_data, expected_keys, conf_file_redaction_file)
    assert err
    assert 'Unknown section' in msg


def test_correct_format_bad_keys_in_wrong_file():
    '''
    Verify that an otherwise valid key is not
    specified in the wrong file (i.e. patterns
    in file-redaction.yaml)
    '''
    parsed_data = {'files': ['/etc/example'], 'patterns': ['abc', 'def']}
    expected_keys = ('files', 'commands')
    err, msg = correct_format(parsed_data, expected_keys, conf_file_redaction_file)
    assert err
    assert 'Unknown section(s) in ' + conf_file_redaction_file in msg


def test_correct_format_bad_invalidtypes():
    '''
    Verify that a config with valid keys,
    but invalid data types, is not allowed
    '''
    parsed_data = {'commands': 'somestring', 'files': ['/var/lib/aaa', '/var/lib/bbb']}
    expected_keys = ('commands', 'files')
    err, msg = correct_format(parsed_data, expected_keys, conf_file_redaction_file)
    assert err
    assert 'must be a list of strings' in msg


def test_correct_format_bad_patterns_keysnoregex():
    '''
    Verify that a config with patterns, if a dict
    with a single key, only contains the key
    "regex"
    '''
    parsed_data = {'patterns': {'wrongkey': ['a(bc)', 'nextregex']}}
    expected_keys = ('patterns', 'keywords')
    err, msg = correct_format(parsed_data, expected_keys, conf_file_content_redaction_file)
    assert err
    assert 'contains an object but the "regex" key was not specified' in msg


def test_correct_format_bad_patterns_invalidkey():
    '''
    Verify that a config with patterns, if a dict
    containing the key "regex", only contains the key "regex"
    '''
    parsed_data = {'patterns': {'regex': [], 'wrongkey': ['a(bc)', 'nextregex']}}
    expected_keys = ('patterns', 'keywords')
    err, msg = correct_format(parsed_data, expected_keys, conf_file_content_redaction_file)
    assert err
    assert 'Only "regex" is valid' in msg


def test_correct_format_bad_patterns_regexinvalidtype():
    '''
    Verify that if a regex key exists in the
    patterns section, that the value is a list
    of strings
    '''
    parsed_data = {'patterns': {'regex': 'a(b)'}}
    expected_keys = ('patterns', 'keywords')
    err, msg = correct_format(parsed_data, expected_keys, conf_file_content_redaction_file)
    assert err
    assert 'regex section under patterns must be a list of strings' in msg


def test_load_yaml_ok():
    '''
    Verify that proper YAML is parsed correctly
    '''
    yaml_data = '---\ncommands:\n- /bin/abc/def\n- /bin/ghi/jkl\nfiles:\n- /etc/abc/def.conf\n'
    with patch_open(yaml_data):
        result = load_yaml('test')
    assert result


def test_load_yaml_error():
    '''
    Verify that improper YAML raises an error
    '''
    yaml_data = '---\ncommands: files:\n- /etc/abc/def.conf\n'
    with patch_open(yaml_data):
        with pytest.raises(RuntimeError) as e:
            result = load_yaml('test')
            assert not result
    assert 'Cannot parse' in str(e.value)


def test_load_yaml_inline_tokens_in_regex_quotes():
    '''
    Verify that, if specifying a regex containing tokens parseable
    by YAML (such as []), when wrapped in quotation marks,
    the regex is parsed properly.
    '''
    yaml_data = '---\npatterns:\n  regex:\n  - \"[[:digit:]]*\"\n'
    with patch_open(yaml_data):
        result = load_yaml('test')
    assert result


def test_load_yaml_inline_tokens_in_regex_noquotes():
    '''
    Verify that, if specifying a regex containing tokens parseable
    by YAML (such as []), when not wrapped in quotation marks,
    an error is raised.
    '''
    yaml_data = '---\npatterns:\n  regex:\n  - [[:digit:]]*\n'
    with patch_open(yaml_data):
        with pytest.raises(RuntimeError) as e:
            result = load_yaml('test')
            assert not result
    assert 'Cannot parse' in str(e.value)


@patch('insights.client.collection_rules.stat.S_IMODE', return_value=0o600)
@patch('insights.client.collection_rules.os.stat', return_value=Mock(st_mode=1))
def test_verify_permissions_ok(os_stat, s_imode):
    '''
    Verify that file permissions 600 does not raise an error
    '''
    verify_permissions('test')


@patch('insights.client.collection_rules.stat.S_IMODE', return_value=0o644)
@patch('insights.client.collection_rules.os.stat', return_value=Mock(st_mode=1))
def test_verify_permissions_bad(os_stat, s_imode):
    '''
    Verify that file permissions 600 does not raise an error
    '''
    with pytest.raises(RuntimeError) as e:
        verify_permissions('test')
    assert 'Invalid permissions' in str(e.value)


# @patch_isfile(True)
# def test_config_filtering(isfile):
#     '''
#     Verify that keys with None values
#     do not appear in the final conf
#     '''
#     filedata = '---\npatterns:\nfiles:\n- /var/lib/aaa'
#     with patch_open(filedata):
#         upload_conf = insights_upload_conf(remove_file=conf_remove_file)
#         result = upload_conf.get_rm_conf()
#     assert 'patterns' not in result and 'files' in result


def patch_isfile(isfile):
    """
    Makes isfile return the passed result.
    """

    def decorator(old_function):
        patcher = patch("insights.client.collection_rules.os.path.isfile", return_value=isfile)
        return patcher(old_function)

    return decorator


def patch_raw_config_parser(items):
    """
    Mocks RawConfigParser so it returns the passed items.
    """

    def decorator(old_function):
        patcher = patch(
            "insights.client.collection_rules.ConfigParser.RawConfigParser",
            **{"return_value.items.return_value": items}
        )
        return patcher(old_function)

    return decorator


@patch_isfile(False)
def test_rm_conf_old_nofile(isfile):
    '''
    Ensure an empty blacklist is generated when the file
    remove.conf does not exist.
    '''
    upload_conf = insights_upload_conf(remove_file=conf_remove_file)
    result = upload_conf.get_rm_conf_old()

    isfile.assert_called_once_with(conf_remove_file)

    # no file, no call to open
    with patch_open('') as mock_open:
        mock_open.assert_not_called()

    assert result is None


@pytest.mark.skipif(mock.version_info < (3, 0, 5), reason="Old mock_open has no iteration control")
@patch('insights.client.collection_rules.verify_permissions', return_value=True)
@patch_isfile(True)
def test_rm_conf_old_emptyfile(isfile, verify):
    '''
    Ensure an empty blacklist is generated when the old
    remove.conf exists, but is empty.
    '''
    filedata = ''
    with patch_open(filedata):
        upload_conf = insights_upload_conf(remove_file=conf_remove_file)
        result = upload_conf.get_rm_conf_old()
    assert result is None


@pytest.mark.skipif(mock.version_info < (3, 0, 5), reason="Old mock_open has no iteration control")
@patch('insights.client.collection_rules.verify_permissions', return_value=True)
@patch_isfile(True)
def test_rm_conf_old_load_bad_invalidsection(isfile, verify):
    '''
    Ensure an error is raised when an invalid
    section is defined in the old remove.conf
    '''
    filedata = '[wrong]\ncommands=/bin/abc'
    with patch_open(filedata):
        upload_conf = insights_upload_conf(remove_file=conf_remove_file)
        with pytest.raises(RuntimeError) as e:
            upload_conf.get_rm_conf_old()
    assert 'ERROR: invalid section(s)' in str(e.value)


@pytest.mark.skipif(mock.version_info < (3, 0, 5), reason="Old mock_open has no iteration control")
@patch('insights.client.collection_rules.verify_permissions', return_value=True)
@patch_isfile(True)
def test_rm_conf_old_load_bad_keysnosection(isfile, verify):
    '''
    Ensure an error is raised when keys are defined without
    a section in the old remove.conf
    '''
    filedata = 'commands=/bin/abc\nfiles=/etc/def'
    with patch_open(filedata):
        upload_conf = insights_upload_conf(remove_file=conf_remove_file)
        with pytest.raises(RuntimeError) as e:
            upload_conf.get_rm_conf_old()
    assert 'ERROR: Cannot parse' in str(e.value)


@pytest.mark.skipif(mock.version_info < (3, 0, 5), reason="Old mock_open has no iteration control")
@patch('insights.client.collection_rules.verify_permissions', return_value=True)
@patch_isfile(True)
def test_rm_conf_old_load_bad_invalidkey(isfile, verify):
    '''
    Ensure an error is raised when an invalid key is defined
    '''
    filedata = '[remove]\ncommands=/bin/abc\nbooradley=/etc/def'
    with patch_open(filedata):
        upload_conf = insights_upload_conf(remove_file=conf_remove_file)
        with pytest.raises(RuntimeError) as e:
            upload_conf.get_rm_conf_old()
    assert 'ERROR: Unknown key' in str(e.value)


@pytest.mark.skipif(mock.version_info < (3, 0, 5), reason="Old mock_open has no iteration control")
@patch('insights.client.collection_rules.verify_permissions', return_value=True)
@patch_isfile(True)
def test_rm_conf_old_load_ok(isfile, verify):
    '''
    Ensure that the old rm_conf load works
    with valid data.
    '''
    filedata = '[remove]\ncommands=/bin/ls,ethtool_i\nfiles=/etc/test\npatterns=abc123,def456\nkeywords=key1,key2,key3'
    with patch_open(filedata):
        upload_conf = insights_upload_conf(remove_file=conf_remove_file)
        result = upload_conf.get_rm_conf_old()
    assert result["commands"] == ["/bin/ls", "ethtool_i"]
    assert result["files"] == ["/etc/test"]
    assert result["patterns"] == ["abc123", "def456"]
    assert result["keywords"] == ["key1", "key2", "key3"]


@patch(
    'insights.client.collection_rules.InsightsUploadConf.load_redaction_file',
    return_value={"files": "/etc/insights-client/machine-id"},
)
@patch("insights.client.collection_rules.logger.warning")
def test_rm_conf_skipped_machine_id_files(warning, config):
    '''
    Verify that '/etc/insights-client/machine-id' is set in file-redaction.yaml[files]
    '''
    upload_conf = insights_upload_conf()
    upload_conf.get_rm_conf()
    warning.assert_called_once_with(
        "WARNING: Spec %s will be skipped for redaction; as it would cause issues, please remove it from %s.",
        "machine_id",
        "/etc/insights-client/file-redaction.yaml",
    )


@patch(
    'insights.client.collection_rules.InsightsUploadConf.load_redaction_file',
    return_value={"files": "machine_id"},
)
@patch("insights.client.collection_rules.logger.warning")
def test_rm_conf_skipped_machine_id_files_symname(warning, config):
    '''
    Verify that 'machine_id' is set in file-redaction.yaml[files]
    '''
    upload_conf = insights_upload_conf()
    upload_conf.get_rm_conf()
    warning.assert_called_once_with(
        "WARNING: Spec %s will be skipped for redaction; as it would cause issues, please remove it from %s.",
        "machine_id",
        "/etc/insights-client/file-redaction.yaml",
    )


@patch(
    'insights.client.collection_rules.InsightsUploadConf.load_redaction_file',
    return_value={"components": "insights.specs.default.DefaultSpecs.machine_id"},
)
@patch("insights.client.collection_rules.logger.warning")
def test_rm_conf_skipped_machine_id_components(warning, config):
    '''
    Verify that 'insights.specs.default.DefaultSpecs.machine_id' is set in file-redaction.yaml[components]
    '''
    upload_conf = insights_upload_conf()
    upload_conf.get_rm_conf()
    warning.assert_called_once_with(
        "WARNING: Spec %s will be skipped for redaction; as it would cause issues, please remove it from %s.",
        "machine_id",
        "/etc/insights-client/file-redaction.yaml",
    )


@patch(
    'insights.client.collection_rules.InsightsUploadConf.load_redaction_file',
    return_value={"commands": "/usr/sbin/subscription-manager identity"},
)
@patch("insights.client.collection_rules.logger.warning")
def test_rm_conf_skipped_subman_id_commands(warning, config):
    '''
    Verify that '/usr/sbin/subscription-manager identity' is set in file-redaction.yaml[commands]
    '''
    upload_conf = insights_upload_conf()
    upload_conf.get_rm_conf()
    warning.assert_called_once_with(
        "WARNING: Spec %s will be skipped for redaction; as it would cause issues, please remove it from %s.",
        "subscription_manager_id",
        "/etc/insights-client/file-redaction.yaml",
    )


@patch(
    'insights.client.collection_rules.InsightsUploadConf.load_redaction_file',
    return_value={"commands": "subscription_manager_id"},
)
@patch("insights.client.collection_rules.logger.warning")
def test_rm_conf_skipped_subman_id_commands_symname(warning, config):
    '''
    Verify that 'subscription_manager_id' is set in file-redaction.yaml[commands]
    '''
    upload_conf = insights_upload_conf()
    upload_conf.get_rm_conf()
    warning.assert_called_once_with(
        "WARNING: Spec %s will be skipped for redaction; as it would cause issues, please remove it from %s.",
        "subscription_manager_id",
        "/etc/insights-client/file-redaction.yaml",
    )


@patch(
    'insights.client.collection_rules.InsightsUploadConf.load_redaction_file',
    return_value={"components": "insights.specs.default.DefaultSpecs.subscription_manager_id"},
)
@patch("insights.client.collection_rules.logger.warning")
def test_rm_conf_skipped_subman_id_components(warning, config):
    '''
    Verify that 'insights.specs.default.DefaultSpecs.subscription_manager_id' is set in file-redaction.yaml[components]
    '''
    upload_conf = insights_upload_conf()
    upload_conf.get_rm_conf()
    warning.assert_called_once_with(
        "WARNING: Spec %s will be skipped for redaction; as it would cause issues, please remove it from %s.",
        "subscription_manager_id",
        "/etc/insights-client/file-redaction.yaml",
    )


@patch(
    'insights.client.collection_rules.InsightsUploadConf.load_redaction_file',
    return_value={"commands": "/usr/sbin/subscription-manager identity", "files": "machine_id"},
)
@patch("insights.client.collection_rules.logger.warning")
def test_rm_conf_skipped_both_id(warning, config):
    '''
    Verify that 'insights.specs.default.DefaultSpecs.subscription_manager_id' is set in file-redaction.yaml[components]
    and 'machine_id' is set in file-redaction.yaml[files]
    '''
    upload_conf = insights_upload_conf()
    upload_conf.get_rm_conf()
    calls = [
        call(
            "WARNING: Spec %s will be skipped for redaction; as it would cause issues, please remove it from %s.",
            "machine_id",
            "/etc/insights-client/file-redaction.yaml",
        ),
        call(
            "WARNING: Spec %s will be skipped for redaction; as it would cause issues, please remove it from %s.",
            "subscription_manager_id",
            "/etc/insights-client/file-redaction.yaml",
        ),
    ]
    warning.assert_has_calls(calls, any_order=True)


# @patch('insights.client.collection_rules.verify_permissions', return_value=True)
# @patch_isfile(True)
# def test_rm_conf_old_load_bad(isfile, verify):
#     '''
#     Ensure that the old rm_conf load rejects
#     invalid data.
#     '''
#     filedata = '[remove]\ncommands=/bin/ls,ethtool_i\nfiles=/etc/test\npatterns=abc123,def456\nkeywords=key1,key2,key3'
#     with patch_open(filedata):
#         upload_conf = insights_upload_conf(remove_file=conf_remove_file)
#         result = upload_conf.get_rm_conf_old()
#     assert result == {'commands': ['/bin/ls', 'ethtool_i'], 'files': ['/etc/test'], 'patterns': ['abc123', 'def456'], 'keywords': ['key1', 'key2', 'key3']}


# @patch('insights.client.collection_rules.InsightsUploadConf.get_rm_conf_old')
# @patch_isfile(True)
# def test_return(isfile, get_rm_conf_old):
#     '''
#     Test that loading YAML from a file will return a dict
#     '''
#     filedata = '---\ncommands:\n- /bin/ls\n- ethtool_i'
#     with patch_open(filedata):
#         upload_conf = insights_upload_conf(remove_file=conf_remove_file)
#         result = upload_conf.get_rm_conf()
#     assert result == {'commands': ['/bin/ls', 'ethtool_i']}
#     get_rm_conf_old.assert_not_called()


# @patch('insights.client.collection_rules.InsightsUploadConf.get_rm_conf_old')
# @patch_isfile(True)
# def test_fallback_to_old(isfile, get_rm_conf_old):
#     '''
#     Test that the YAML function falls back to classic INI
#     if the file cannot be parsed as YAML
#     '''
#     filedata = 'ncommands\n /badwain/ls\n- ethtool_i'
#     with patch_open(filedata):
#         upload_conf = insights_upload_conf(remove_file=conf_remove_file)
#         upload_conf.get_rm_conf()
#     get_rm_conf_old.assert_called_once()


# @pytest.mark.skipif(mock.version_info < (3, 0, 5), reason="Old mock_open has no iteration control")
# @patch_isfile(True)
# def test_fallback_ini_data(isfile):
#     '''
#     Test that the YAML function falls back to classic INI
#     if the file cannot be parsed as YAML, and the data is
#     parsed as INI
#     '''
#     filedata = '[remove]\ncommands=/bin/ls,ethtool_i'
#     with patch_open(filedata):
#         upload_conf = insights_upload_conf(remove_file=conf_remove_file)
#         result = upload_conf.get_rm_conf()
#     assert result == {'commands': ['/bin/ls', 'ethtool_i']}


# @pytest.mark.skipif(mock.version_info < (3, 0, 5), reason="Old mock_open has no iteration control")
# @patch_isfile(True)
# def test_fallback_bad_data(isfile):
#     '''
#     Test that the YAML function falls back to classic INI
#     if the file cannot be parsed as YAML, and the data isn't
#     INI either so it's thrown out
#     '''
#     filedata = 'ncommands\n /badwain/ls\n- ethtool_i'
#     with patch_open(filedata):
#         upload_conf = insights_upload_conf(remove_file=conf_remove_file)
#         with pytest.raises(RuntimeError) as e:
#             upload_conf.get_rm_conf()
#     assert 'YAML file nor as an INI file' in str(e.value)


# @patch_isfile(True)
# def test_load_string_patterns(isfile):
#     '''
#     Test that the patterns section is loaded as a list of strings.
#     '''
#     filedata = '---\npatterns:\n- abcd\n- bcdef'
#     with patch_open(filedata):
#         upload_conf = insights_upload_conf(remove_file=conf_remove_file)
#         result = upload_conf.get_rm_conf()
#     assert 'patterns' in result
#     assert isinstance(result['patterns'], list)


# @patch_isfile(True)
# def test_load_string_regex(isfile):
#     '''
#     Test that the patterns section is loaded as a dict with
#     key 'regex' and the value is a list of strings
#     '''
#     filedata = '---\npatterns:\n  regex:\n  - abcd\n  - bcdef'
#     with patch_open(filedata):
#         upload_conf = insights_upload_conf(remove_file=conf_remove_file)
#         result = upload_conf.get_rm_conf()
#     assert 'patterns' in result
#     assert isinstance(result['patterns'], dict)
#     assert 'regex' in result['patterns']
#     assert isinstance(result['patterns']['regex'], list)


# @patch_raw_config_parser([("files", ",".join(removed_files))])
# @patch_isfile(True)
# def test_return_old(isfile, raw_config_parser):
#     upload_conf = insights_upload_conf(remove_file=conf_remove_file)
#     result = upload_conf.get_rm_conf_old()

#     raw_config_parser.assert_called_once_with()
#     raw_config_parser.return_value.read.assert_called_with(conf_remove_file)
#     raw_config_parser.return_value.items.assert_called_with('remove')

#     assert result == {"files": removed_files}


# def test_raw_config_parser():
#     '''
#         Ensure that get_rm_conf and json.loads (used to load uploader.json) return the same filename
#     '''
#     raw_filename = '/etc/yum/pluginconf.d/()*\\\\w+\\\\.conf'
#     uploader_snip = json.loads('{"pattern": [], "symbolic_name": "pluginconf_d", "file": "' + raw_filename + '"}')
#     with open(conf_remove_file, 'w') as rm_conf:
#         rm_conf.write('[remove]\nfiles=' + raw_filename)
#     coll = InsightsUploadConf(InsightsConfig(remove_file=conf_remove_file))
#     items = coll.get_rm_conf()
#     assert items['files'][0] == uploader_snip['file']


# @patch_isfile(True)
# def test_config_filtering(isfile):
#     '''
#     Verify that keys with None values
#     do not appear in the final conf
#     '''
#     filedata = '---\npatterns:\nfiles:\n- /var/lib/aaa'
#     with patch_open(filedata):
#         upload_conf = insights_upload_conf(remove_file=conf_remove_file)
#         result = upload_conf.get_rm_conf()
#     assert 'patterns' not in result and 'files' in result



# -*- coding: UTF-8 -*-
# flake8: noqa: E402
import os
import collections
import pkgutil
import sys

import mock
import pytest
import tempfile
import shutil
from mock.mock import patch
from pytest import raises

from insights.client.constants import InsightsConstants as constants
from insights.client.apps.ansible import playbook_verifier
from insights.client.apps.ansible.playbook_verifier import verify, PlaybookVerificationError, get_play_revocation_list, normalize_play_py2, load_playbook_yaml  # noqa


SKIP_ON_3 = pytest.mark.skipif(sys.version_info[0] > 2, reason="Only required in Python 2")


class TestErrors:
    @patch("insights.client.apps.ansible.playbook_verifier.get_play_revocation_list", return_value=[])
    def test_vars_not_found_error(self, mock_method):
        vars_error = "Play doesn't have a section 'vars'."
        fake_playbook = {'name': "test playbook"}

        with raises(PlaybookVerificationError) as error:
            verify(fake_playbook)
        assert vars_error in str(error.value)

    @patch("insights.client.apps.ansible.playbook_verifier.get_play_revocation_list", return_value=[])
    def test_empty_vars_error(self, mock_method):
        sig_error = "Play doesn't have a section 'vars'."
        fake_playbook = {'name': "test playbook", 'vars': None}

        with raises(PlaybookVerificationError) as error:
            verify(fake_playbook)
        assert sig_error in str(error.value)

    @patch("insights.client.apps.ansible.playbook_verifier.get_play_revocation_list", return_value=[])
    def test_signature_not_found_error(self, mock_method):
        sig_error = "Play doesn't contain the Insights signature."
        fake_playbook = {'name': "test playbook", 'vars': {}}

        with raises(PlaybookVerificationError) as error:
            verify(fake_playbook)
        assert sig_error in str(error.value)

    @patch('insights.client.apps.ansible.playbook_verifier.PUBLIC_KEY_PATH', './testing')
    def test_key_not_imported(self):
        key_error = "Public key import failed."
        fake_playbook = {
            'name': "test playbook",
            'vars': {
                'insights_signature': 'TFMwdExTMUNSVWRKVGlCUVIxQWdVMGxIVGtGVVZWSkZMUzB0TFMwS0N==',
                'insights_signature_exclude': '/vars/insights_signature'
            }
        }

        # Create a temporary directory as the mocked insights_core_lib_dir
        # in insights constants instead of directly manipulating
        # the system insights lib directory.
        temp_dir = tempfile.mkdtemp(dir="/tmp/", prefix="insights_")
        try:
            with patch.object(constants, "insights_core_lib_dir", temp_dir):
                with raises(PlaybookVerificationError) as error:
                    verify(fake_playbook)
                assert key_error in str(error.value)
        finally:
            if os.path.exists(temp_dir):
                try:
                    shutil.rmtree(temp_dir)
                except OSError:
                    pass

    @patch('insights.client.apps.ansible.playbook_verifier.PUBLIC_KEY_PATH', None)
    def test_key_import_error(self):
        key_error = "Public key file not found."
        fake_playbook = {
            'name': "test playbook",
            'vars': {
                'insights_signature': 'TFMwdExTMUNSVWRKVGlCUVIxQWdVMGxIVGtGVVZWSkZMUzB0TFMwS0N==',
                'insights_signature_exclude': '/vars/insights_signature'
            }
        }

        # Create a temporary directory as the mocked insights_core_lib_dir
        # in insights constants instead of directly manipulating
        # the system insights lib directory.
        temp_dir = tempfile.mkdtemp(dir="/tmp/", prefix="insights_")
        try:
            with patch.object(constants, "insights_core_lib_dir", temp_dir):
                with raises(PlaybookVerificationError) as error:
                    verify(fake_playbook)
                assert key_error in str(error.value)
        finally:
            if os.path.exists(temp_dir):
                try:
                    shutil.rmtree(temp_dir)
                except OSError:
                    pass

    @patch('insights.client.apps.ansible.playbook_verifier.verify_play', return_value=([], []))
    @patch('insights.client.apps.ansible.playbook_verifier.get_play_revocation_list', return_value=[])
    def test_playbook_verification_error(self, call_1, call_2):
        key_error = "Play 'test playbook' has invalid signature"
        fake_playbook = {
            'name': "test playbook",
            'vars': {
                'insights_signature': 'TFMwdExTMUNSVWRKVGlCUVIxQWdVMGxIVGtGVVZWSkZMUzB0TFMwS0N==',
                'insights_signature_exclude': '/vars/insights_signature'
            }
        }

        with raises(PlaybookVerificationError) as error:
            verify(fake_playbook)
        assert key_error in str(error.value)

    @patch('insights.client.apps.ansible.playbook_verifier.contrib.gnupg.GPG.verify_data')
    def test_playbook_verification_success(self, mock_method):
        mock_method.return_value = mock.MagicMock(valid=True, status="mocked status")
        fake_playbook = {
            'name': "test playbook",
            'vars': {
                'insights_signature': 'TFMwdExTMUNSVWRKVGlCUVIxQWdVMGxIVGtGVVZWSkZMUzB0TFMwS0N==',
                'insights_signature_exclude': '/vars/insights_signature'
            }
        }

        # Create a temporary directory as the mocked insights_core_lib_dir
        # in insights constants instead of directly manipulating
        # the system insights lib directory.
        temp_dir = tempfile.mkdtemp(dir="/tmp/", prefix="insights_")
        try:
            with patch.object(constants, "insights_core_lib_dir", temp_dir):
                result = verify(fake_playbook)
                assert result == fake_playbook
        finally:
            if os.path.exists(temp_dir):
                try:
                    shutil.rmtree(temp_dir)
                except OSError:
                    pass

    # get_play_revocation_list can't load list
    @patch('insights.client.apps.ansible.playbook_verifier.contrib.ruamel_yaml.ruamel.yaml.YAML.load', side_effect=Exception())
    def test_revocation_list_not_found(self, mock_method):
        load_error = "Could not load play revocation list."

        with raises(PlaybookVerificationError) as error:
            get_play_revocation_list(pkgutil.get_data('insights', 'revoked_playbooks.yaml'))

        assert load_error in str(error.value)

    # revocation list signature invalid
    @patch('insights.client.apps.ansible.playbook_verifier.verify_play', return_value=(None, 0xdeadbeef))
    def test_revocation_list_signature_invalid(self, mock_method):
        load_error = "List of revocation signatures is invalid."

        with raises(PlaybookVerificationError) as error:
            get_play_revocation_list(pkgutil.get_data('insights', 'revoked_playbooks.yaml'))

        assert load_error in str(error.value)

    # revocation list empty
    @patch('insights.client.apps.ansible.playbook_verifier.contrib.ruamel_yaml.ruamel.yaml.YAML.load', return_value=[{}])
    @patch('insights.client.apps.ansible.playbook_verifier.verify_play', return_value=(True, 0xdeadbeef))
    def test_revocation_list_empty(self, call_1, call_2):
        fake_playbook = {
            'name': "test playbook",
            'vars': {
                'insights_signature': 'TFMwdExTMUNSVWRKVGlCUVIxQWdVMGxIVGtGVVZWSkZMUzB0TFMwS0N==',
                'insights_signature_exclude': '/vars/insights_signature'
            }
        }

        result = verify(fake_playbook)
        assert result == fake_playbook

    # playbook on revoked list
    @patch('insights.client.apps.ansible.playbook_verifier.contrib.ruamel_yaml.ruamel.yaml.YAML.load',
           return_value=[{'revoked_playbooks': [{'name': 'banned book', 'hash': 'deadbeef'}]}])
    @patch('insights.client.apps.ansible.playbook_verifier.verify_play', return_value=(True, bytearray.fromhex(u'deadbeef')))
    def test_revoked_playbook(self, call_1, call_2):
        revoked_error = "Play 'test playbook' is on the revoked list."
        fake_playbook = {
            'name': "test playbook",
            'vars': {
                'insights_signature': 'TFMwdExTMUNSVWRKVGlCUVIxQWdVMGxIVGtGVVZWSkZMUzB0TFMwS0N==',
                'insights_signature_exclude': '/vars/insights_signature'
            }
        }

        with raises(PlaybookVerificationError) as error:
            verify(fake_playbook)

        assert revoked_error in str(error.value)


@SKIP_ON_3
def test_normalize_snippet():
    playbook = '''task:
  when:
    - '"pam" in ansible_facts.packages'
    - result_pam_file_present.stat.exists'''

    snippet = load_playbook_yaml(playbook)

    want = {
        'task': {
            'when': [
                '"pam" in ansible_facts.packages',
                'result_pam_file_present.stat.exists'
            ]
        }
    }

    assert normalize_play_py2(snippet) == want


class TestExcludeDynamicElements:
    def test_ok_signature(self):
        source = {
            "vars": {"insights_signature_exclude": "/vars/insights_signature", "insights_signature": b"binary data"}
        }
        expected = {"vars": {"insights_signature_exclude": "/vars/insights_signature"}}
        result = playbook_verifier.exclude_dynamic_elements(source)
        assert result == expected

    def test_ok_hosts(self):
        source = {"vars": {"insights_signature_exclude": "/hosts"}, "hosts": "localhost"}
        expected = {"vars": {"insights_signature_exclude": "/hosts"}}
        result = playbook_verifier.exclude_dynamic_elements(source)
        assert result == expected

    def test_ok_signature_and_hosts(self):
        source = {
            "vars": {"insights_signature_exclude": "/vars/insights_signature,/hosts", "insights_signature": b""},
            "hosts": "localhost",
        }
        expected = {"vars": {"insights_signature_exclude": "/vars/insights_signature,/hosts"}}
        result = playbook_verifier.exclude_dynamic_elements(source)
        assert result == expected

    def test_ok_vars_foo_bar(self):
        source = {"vars": {"insights_signature_exclude": "/vars/foo,/vars/bar", "foo": "foo", "bar": "bar"}}
        expected = {"vars": {"insights_signature_exclude": "/vars/foo,/vars/bar"}}
        result = playbook_verifier.exclude_dynamic_elements(source)
        assert result == expected

    def test_ok_trailing_slash(self):
        source = {"vars": {"insights_signature_exclude": "/vars/insights_signature/", "insights_signature": b""}}
        expected = {"vars": {"insights_signature_exclude": "/vars/insights_signature/"}}
        result = playbook_verifier.exclude_dynamic_elements(source)
        assert result == expected

    def test_missing_vars(self):
        source = {}
        expected = (
            "Play does not have the key 'vars/insights_signature_exclude', "
            "dynamic exclusion cannot be performed."
        )
        with pytest.raises(PlaybookVerificationError) as excinfo:
            playbook_verifier.exclude_dynamic_elements(source)
        assert excinfo.value.message == expected

    def test_missing_vars_exclude(self):
        source = {"vars": {"insights_signature": b""}}
        expected = (
            "Play does not have the key 'vars/insights_signature_exclude', "
            "dynamic exclusion cannot be performed."
        )
        with pytest.raises(PlaybookVerificationError) as excinfo:
            playbook_verifier.exclude_dynamic_elements(source)
        assert excinfo.value.message == expected

    def test_missing_exclusion_key_in_vars(self):
        source = {"vars": {"insights_signature_exclude": "/vars/missing"}}
        expected = "Dynamic key '/vars/missing' does not exist and could not be excluded."
        with pytest.raises(PlaybookVerificationError) as excinfo:
            playbook_verifier.exclude_dynamic_elements(source)
        assert excinfo.value.message == expected

    def test_missing_exclusion_key_in_root(self):
        source = {"vars": {"insights_signature_exclude": "/hosts"}}
        expected = "Dynamic key '/hosts' does not exist and could not be excluded."
        with pytest.raises(PlaybookVerificationError) as excinfo:
            playbook_verifier.exclude_dynamic_elements(source)
        assert excinfo.value.message == expected

    def test_deep_exclusion_key(self):
        source = {"vars": {"insights_signature_exclude": "/vars/too/deep"}, "too": {"deep": ""}}
        expected = "Dynamic key '/vars/too/deep' cannot be excluded from verification."
        with pytest.raises(PlaybookVerificationError) as excinfo:
            playbook_verifier.exclude_dynamic_elements(source)
        assert excinfo.value.message == expected

    def test_missing_deep_exclusion_key(self):
        source = {"vars": {"insights_signature_exclude": "/vars/too/deep"}}
        expected = "Dynamic key '/vars/too/deep' cannot be excluded from verification."
        with pytest.raises(PlaybookVerificationError) as excinfo:
            playbook_verifier.exclude_dynamic_elements(source)
        assert excinfo.value.message == expected

    def test_invalid_exclusion_key(self):
        source = {"name": "test", "vars": {"insights_signature_exclude": "/name"}}
        expected = "Dynamic key '/name' cannot be excluded from verification."
        with pytest.raises(PlaybookVerificationError) as excinfo:
            playbook_verifier.exclude_dynamic_elements(source)
        assert excinfo.value.message == expected


class TestPlaybookSerializer:

    def test_list(self):
        source = ["value1", "value2"]
        result = playbook_verifier.PlaybookSerializer.serialize(source)
        expected = "['value1', 'value2']"
        assert result == expected

    def test_dict_empty(self):
        source = {}
        result = playbook_verifier.PlaybookSerializer.serialize(source)
        expected = "ordereddict()"
        assert result == expected

    def test_dict_empty_value(self):
        source = {"key": None}
        result = playbook_verifier.PlaybookSerializer.serialize(source)
        expected = "ordereddict([('key', None)])"
        assert result == expected

    def test_dict_single_key(self):
        source = {"key": "value"}
        result = playbook_verifier.PlaybookSerializer.serialize(source)
        expected = "ordereddict([('key', 'value')])"
        assert result == expected

    def test_dict_value_list(self):
        source = {"key": ["value1", "value2"]}
        result = playbook_verifier.PlaybookSerializer.serialize(source)
        expected = "ordereddict([('key', ['value1', 'value2'])])"
        assert result == expected

    # Python 2.7 dictionaries do not keep insertion order, we have to use OrderedDicts.

    def test_dict_mixed_value_types(self):
        source = collections.OrderedDict([("key", "key"), ("value", ["value1", "value2"])])
        result = playbook_verifier.PlaybookSerializer.serialize(source)
        expected = "ordereddict([('key', 'key'), ('value', ['value1', 'value2'])])"
        assert result == expected

    def test_dict_multiple_keys(self):
        source = collections.OrderedDict([("key", "key"), ("value", "value")])
        result = playbook_verifier.PlaybookSerializer.serialize(source)
        expected = "ordereddict([('key', 'key'), ('value', 'value')])"
        assert result == expected

    @pytest.mark.parametrize("source,expected", [(37, "37"), (17.93233901, "17.93233901")])
    def test_numbers(self, source, expected):
        result = playbook_verifier.PlaybookSerializer.serialize(source)
        assert result == expected

    @pytest.mark.parametrize(
        "source,expected",
        [
            ("no quote", "'no quote'"),
            ("single'quote", '''"single'quote"'''),
            ("double\"quote", """'double"quote'"""),
            ("both\"'quotes", r"""'both"\'quotes'"""),
            ("\\backslash", "'\\\\backslash'"),
            ("new\nline", "'new\\nline'"),
            ("tab\tchar", "'tab\\tchar'"),
        ]
    )
    def test_strings(self, source, expected):
        result = playbook_verifier.PlaybookSerializer.serialize(source)
        assert result == expected

    @pytest.mark.parametrize(
        "source,expected",
        [
            pytest.param("zw​space", "'zw\\u200bspace'", id="zero-width space"),
            pytest.param("zw‌nonjoiner", "'zw\\u200cnonjoiner'", id="zero-width non-joiner"),
            pytest.param("👨🏼‍🚀", "'👨🏼\\u200d🚀'", id="zero-width joiner"),
        ],
    )
    def test_strings_unicode(self, source, expected):
        if sys.version_info < (3, 0):
            raise pytest.skip("Unicode characters are not supported on Python 2 systems")

        result = playbook_verifier.PlaybookSerializer.serialize(source)
        assert result == expected


class TestSerializePlaybookSnippet:
    def test_serialize_dictionary(self):
        raw = "\n".join([
            "---",
            "- tasks:",
            "    - name: name",
            "      command: command",
        ])
        expected = (
            b"ordereddict(["
            b"('tasks', [ordereddict(["
            b"('name', 'name'), ('command', 'command')"
            b"])])"
            b"])"
        )
        playbooks = playbook_verifier.load_playbook_yaml(raw)  # type: list[dict]
        result = playbook_verifier.serialize_play(playbooks[0])  # type: bytes
        assert result == expected

    def test_small(self):
        raw_playbook = "\n".join([
            "---",
            '- name: test',
            '  vars:',
            '    insights_signature_exclude: /hosts',
            '  tasks:',
            '    - name: show current date',
            '      command: date | cowsay | lolcat',
        ])
        expected = (
            b"ordereddict(["
            b"('name', 'test'), "
            b"('vars', ordereddict(["
            b"('insights_signature_exclude', '/hosts')"
            b"])), "
            b"('tasks', [ordereddict(["
            b"('name', 'show current date'), ('command', 'date | cowsay | lolcat')"
            b"])])"
            b"])"
        )

        playbooks = playbook_verifier.load_playbook_yaml(raw_playbook)  # type: list[dict]
        result = playbook_verifier.serialize_play(playbooks[0])  # type: bytes
        assert result == expected

    @pytest.mark.parametrize("filename", ("insights_setup", "insights_remove", "document-from-hell", "unicode"))
    def test_real(self, filename):
        if filename == "unicode" and sys.version_info < (3, 7):
            raise pytest.skip("Python 3.7 or later is required to test playbooks containing unicode")

        parent = os.path.dirname(__file__)  # type: str
        with open("{parent}/playbooks/{filename}.yml".format(parent=parent, filename=filename), "r") as f:
            playbooks = load_playbook_yaml(f.read())  # type: list[dict]
        playbook = playbook_verifier.exclude_dynamic_elements(playbooks[0])  # type: dict
        with open("{parent}/playbooks/{filename}.serialized.bin".format(parent=parent, filename=filename), "rb") as f:
            expected = f.read()  # type: bytes

        result = playbook_verifier.serialize_play(playbook)  # type: bytes
        assert result == expected


class TestGetPlaybookSnippetRevocationList:
    @mock.patch(
        "insights.client.apps.ansible.playbook_verifier.verify_play",
        return_value=(True, b"snippet hash"),
    )
    def test_ok(self, mocked_verify_play):
        revoked_playbooks = (
            b"---\n"
            b"- name: revocation list\n"
            b"  timestamp: 123\n"
            b"  vars:\n"
            b"    insights_signature_exclude: /vars/insights_signature\n"
            b"    insights_signature: !!binary |\n"
            b"      content==\n"
            b"  revoked_playbooks:\n"
            b"    - name: a.yml hash8a\n"
            b"      hash: hash64a\n"
            b"    - name: b.yml hash8b\n"
            b"      hash: hash64b\n"
        )
        revocation_list = playbook_verifier.get_play_revocation_list(revoked_playbooks)
        expected = [{"name": "a.yml hash8a", "hash": "hash64a"}, {"name": "b.yml hash8b", "hash": "hash64b"}]

        assert mocked_verify_play.call_count == 1
        assert revocation_list == expected

    def test_malformed(self):
        revoked_playbooks = (
            b"---\n"
            b"- name: foo\n"
            b"  vars:\n"
            b"    key: value\n"
            b"text that should not be here"
        )
        with pytest.raises(playbook_verifier.PlaybookVerificationError) as excinfo:
            playbook_verifier.get_play_revocation_list(revoked_playbooks)
        assert "Could not load" in excinfo.value.message

    @mock.patch(
        "insights.client.apps.ansible.playbook_verifier.verify_play",
        return_value=(False, b"snippet hash"),
    )
    def test_invalid_signature(self, mocked_verify_play):
        revoked_playbooks = (
            b"---\n"
            b"- name: foo\n"
            b"  vars:\n"
            b"    key: value\n"
        )
        with pytest.raises(playbook_verifier.PlaybookVerificationError) as excinfo:
            playbook_verifier.get_play_revocation_list(revoked_playbooks)
        assert "invalid" in excinfo.value.message
        assert mocked_verify_play.call_count == 1


class TestHashPlaybookSnippets:
    @pytest.mark.parametrize("filename", ("insights_remove", "document-from-hell", "unicode"))
    def test_real(self, filename):
        parent = os.path.dirname(__file__)  # type: str
        with open("{parent}/playbooks/{filename}.serialized.bin".format(parent=parent, filename=filename), "rb") as f:
            serialized_playbook = f.read()  # type: bytes

        with open("{parent}/playbooks/{filename}.digest.bin".format(parent=parent, filename=filename), "rb") as f:
            expected = f.read()  # type: bytes

        result = playbook_verifier.hash_play(serialized_playbook)  # type: bytes
        assert result == expected



"""
OSRelease
=========
The ``OSRelease`` combiner uses the following parsers to try to identify if the
current host is installed with a "Red Hat Enterprise Linux" (RHEL) system.

    - :py:class:`insights.parsers.installed_rpms.InstalledRpms`
    - :py:class:`insights.parsers.uname.Uname`
    - :py:class:`insights.parsers.dmesg.DmesgLineList`
    - :py:class:`insights.parsers.os_release.OSRelease`
    - :py:class:`insights.parsers.redhat_release.RedhatRelease`

It provides an attribute `is_rhel` that indicates if the host is RHEL or not.
It also provides an attribute `release` which returns the estimated OS release
name of the system, "Unknown" will be returned by default when cannot identify
the OS.

* TODO::

    The lists of keywords to identify NON-RHEL system of each sub-combiners are
    based on our current knowledge, and may be not sufficient. It needs to be
    updated timely according to new found Linux Distributions.
"""
from insights.core.filters import add_filter
from insights.core.plugins import combiner
# TODO: replace DmesgLineList with '/proc/version' (not collected yet)
from insights.parsers.dmesg import DmesgLineList
from insights.parsers.installed_rpms import InstalledRpm, InstalledRpms
from insights.parsers.os_release import OsRelease
from insights.parsers.redhat_release import RedhatRelease
from insights.parsers.uname import Uname

RHEL_KEYS = ['rhel', 'red hat enterprise linux']
OTHER_LINUX_KEYS = {
    # other_linux: (dmesg-keywords, release-packages)
    'Fedora': (
        ['fedora'],
        ['fedora-release']),
    'CentOS': (
        ['centos'],
        ['centos-stream-release', 'centos-release', 'centos-linux-release']),
    'Oracle': (
        ['oracle'],
        ['enterprise-release', 'oraclelinux-release']),
    'CloudLinux': (
        ['cloudlinux'],
        ['cloudlinux-release']),
    'ClearOS': (
        ['clearos'],
        ['clearos-release']),
    'AlmaLinux': (
        ['almalinux'],
        ['almalinux-release', 'almalinux-logos', 'almalinux-backgrounds']),
    'Rocky': (
        ['rockylinux', 'rocky'],
        ['rocky-release']),
    'Scientific': (
        [],  # Empty
        ['sl-release']),
    'SUSE': (
        ['suse', 'sles', 'novell'],
        ['sles-release', 'sles_es-release-server']),
    'Miracle': (
        [],  # Empty
        ['miraclelinux-release']
    )
}
"""Known NON-RHEL Linux Distributions."""
# TODO:
# Update the CONVERT2RHEL_SUPPORTED list when necessary
CONVERT2RHEL_SUPPORTED = ['CentOS']
# Get "Linux version" from `dmesg`
DMESG_LINUX_BUILD_INFO = 'Linux version'
add_filter(DmesgLineList, DMESG_LINUX_BUILD_INFO)
DmesgLineList.keep_scan("linux_version", DMESG_LINUX_BUILD_INFO, num=1)
# Must-install packages for minimum RHEL system
MINIMUM_RHEL_PKGS = [
    # 'kernel' is checked individually (booting kernel)
    'audit-libs',
    'basesystem',
    'bash',
    'coreutils',
    'dbus',
    'dmidecode',
    'dnf',  # RHEL 8+
    'dracut',
    'filesystem',
    'firewalld',
    'glibc',
    'gmp',
    'krb5-libs',
    'libacl',
    'libgcc',
    'libselinux',
    'NetworkManager',
    'openssl-libs',
    'passwd',
    'redhat-release',  # RHEL 8+
    'redhat-release-server',  # RHEL 6/7
    'systemd',  # RHEL 7+
    'util-linux',
    'yum',  # RHEL 6/7
]
"""Must-install packages for minimum installed RHEL system."""
THRESHOLD = 0.75
"""Threshold of the must-install packages to identify NON-RHEL."""


def _get_release(names):
    names = names if isinstance(names, list) else [names]
    for rel in OTHER_LINUX_KEYS:
        if any(rel.lower() in name.lower() for name in names):
            return rel
    return names[-1].split()[0]


def _from_os_release(osr):
    """
    Identify RHEL by checking the `/etc/os-release`.
    """
    def _filter(name):
        """Remove falsy or items contain RHEL info"""
        if not name or any(k in name.lower() for k in RHEL_KEYS):
            return False
        return name

    if osr:
        names = list(filter(_filter, [osr.get('ID'), osr.get('NAME'),
                                      osr.get('PRETTY_NAME')]))
        if names:
            # NON-RHEL: /etc/os-release
            return dict(other_linux=_get_release(names),
                        reason='NON-RHEL: os-release')
        return dict(other_linux='RHEL')


def _from_redhat_release(rhr):
    """
    Identify RHEL by checking the `/etc/redhat-release`.
    """
    if rhr:
        if not rhr.is_rhel:
            return dict(other_linux=_get_release(rhr.product),
                        reason='NON-RHEL: redhat-release')
        return dict(other_linux='RHEL')


def _from_uname(uname):
    """
    Identify RHEL by checking the `uname -a`.

    1. Oracle kernels may contain 'uek' or 'ol' in the kernel NVR.
    2. Fedora kernels contains 'fc' in the NVR.
    3. RHEL kernels have '.el' in the NVR
       RHEL based Linux kernels may also have the '.el' in the NVR,
       but they are also checked in other sub-combiners.
    4. Otherwise, flag it as an "Unknown"
    """
    LINUX_UNAME_KEYS = [
        ('Oracle', ['uek', 'ol']),
        ('Fedora', ['fc']),
        ('RHEL', ['.el']),  # the last item
    ]
    if uname:
        kernel = uname.kernel
        ret = dict(other_linux='Unknown')
        for rel, keys in LINUX_UNAME_KEYS:
            if any(key in kernel for key in keys):
                ret.update(other_linux=rel)
                break
        if ret.get('other_linux') != 'RHEL':
            ret.update(kernel=kernel)
        return ret


def _from_dmesg(dmesg):
    """
    Identify RHEL by checking the `dmesg`.

    The `dmesg` includes a line containing the kernel build information,
    e.g. the build host and GCC version.
    If this line doesn't contain 'redhat.com' then we can assume the kernel
    wasn't built on a Red Hat machine and this should be flagged.
    """
    if dmesg and dmesg.linux_version:
        line = dmesg.linux_version[0]['raw_message']
        low_line = line.lower()
        if 'redhat.com' not in low_line:
            release = 'Unknown'
            for rel, keys in OTHER_LINUX_KEYS.items():
                if any(kw in low_line for kw in keys[0]):
                    release = rel
                    break
            return dict(other_linux=release, build_info=line)
        else:
            return dict(other_linux='RHEL')


def _from_installed_rpms(rpms, uname):
    """
    Identify RHEL by checking the `installed_rpms`.

    Two parts are included, see below:
    """
    if not rpms:
        return
    # Part-1: the known non-rhel-release packages exists
    for rel, pkgs in OTHER_LINUX_KEYS.items():
        for pkg_name in pkgs[1]:
            pkg = rpms.newest(pkg_name)
            if pkg:
                return dict(other_linux=rel, release=pkg.nvr)
    # Part-2: too many must-install packages are NOT from Red Hat
    # - more than THRESHOLD packages are not signed and not provided by Red Hat
    #   faulty_packages >= THRESHOLD * must-install packages
    installed_packages = 0
    vendor, ng_pkgs = '', set()
    if uname:
        # check the booting 'kernel' first
        installed_packages += 1
        boot_kn = InstalledRpm.from_package('kernel-{0}'.format(uname.kernel))
        for pkg in rpms.packages.get('kernel', []):
            if pkg == boot_kn:
                vendor = pkg.vendor
                if pkg.redhat_signed is False and vendor != 'Red Hat, Inc.':
                    ng_pkgs.add(pkg.nvr)
                # check the booting kernel only
                break
    # check other packages
    for pkg_name in MINIMUM_RHEL_PKGS:
        pkg = rpms.newest(pkg_name)
        if pkg:
            # count the package only when it's installed
            installed_packages += 1
            if pkg.redhat_signed is False and pkg.vendor != 'Red Hat, Inc.':
                ng_pkgs.add(pkg.nvr)
    # check the result
    if len(ng_pkgs) >= round(THRESHOLD * installed_packages) > 0:
        # NON-RHEL: more than THRESHOLD packages are NOT from Red Hat
        ret = dict(other_linux='Unknown', faulty_packages=sorted(ng_pkgs))
        if vendor:
            ret.update(kernel_vendor=vendor)
            # try to get the release from kernel vendor
            if 'red hat' not in vendor.lower():
                sep = ',' if ',' in vendor else ' '
                release = _get_release(vendor.split(sep)[0].strip())
                ret.update(other_linux=release)
        return ret
    return dict(other_linux='RHEL')


@combiner(optional=[Uname, DmesgLineList, InstalledRpms,
                    OsRelease, RedhatRelease])
class OSRelease(object):
    """
    A Combiner identifies whether the current Linux a Red Hat Enterprise Linux
    or not.

    Examples:
        >>> type(osr)
        <class 'insights.combiners.os_release.OSRelease'>
        >>> osr.is_rhel
        False
        >>> osr.release == "Oracle"
        True
        >>> osr.name == "Oracle Linux Server"
        True
        >>> osr.is_rhel_compatible
        False
        >>> 'kernel' in osr.reasons.keys()
        False
        >>> 'faulty_packages' in osr.reasons.keys()
        True
        >>> 'glibc-2.28-211.el8' in osr.reasons['faulty_packages']
        True
    """
    def __init__(self, uname, dmesg, rpms, osr, rhr):

        def __identify_non_rhel(ret):
            if ret and 'other_linux' in ret:
                self._release = ret.pop('other_linux')
                self._reasons = ret
                if self._release == 'Unknown':
                    return None  # Continue to identify when unknown
                return self._release != 'RHEL'  # see OTHER_LINUX_KEYS
            return None  # Nothing to check

        self._release = 'Unknown'
        self._reasons = dict(reason='Nothing available to check')

        # 1. Check `installed_rpms` first.
        ret = _from_installed_rpms(rpms, uname)
        #    We trust the check result of `installed_rpms`: RHEL or NON-RHEL,
        #    expect for "Unknown".
        #    `None` indicates "Unknown" or `installed_rpms` is not available
        if __identify_non_rhel(ret) is None:
            # 2. Only when `installed_rpms` cannot identify it
            #    a. installed_rpms is not available
            #    b. identify result is "Unknown"
            #
            #    Check below items with order:
            #    - uname, dmesg, os_release, and redhat_release
            #    and any `False` which indicates NON-RHEL to stop.
            check_points_funcs = [
                (_from_uname, [uname]),
                (_from_dmesg, [dmesg]),
                (_from_os_release, [osr]),
                (_from_redhat_release, [rhr]),
            ]

            for chk_func, args in check_points_funcs:
                if __identify_non_rhel(chk_func(*args)):
                    break

        self._name = osr.get('NAME', self._release) if osr else self._release

    @property
    def is_rhel(self):
        """
        Returns True if it's RHEL, False for NON-RHEL.
        """
        return self._release == 'RHEL'

    @property
    def is_rhel_compatible(self):
        """
        Returns True if Convert2RHEL could convert
        """
        return self._release in CONVERT2RHEL_SUPPORTED

    @property
    def release(self):
        """
        Returns the estimated release name of the running Linux.  It's RHEL or
        one key of the :const:`OTHER_LINUX_KEYS` when it's NON-RHEL.
        """
        return self._release

    @property
    def name(self):
        """
        Returns the name of the OS. Generally it's the ``NAME`` of the
        `/etc/os-release` file, or it's the same as the `release` when
        the `/etc/os-release` is not available.
        """
        return self._name

    @property
    def reasons(self):
        """
        Returns a dict indicating why the host is a NON-RHEL.  Empty when
        it's an RHEL. The keys include::

            kernel (str): the kernel package
            build_info (str): the kernel build information
            release (str): the release package
            faulty_packages (list): the packages that are not signed and not
                provided by Red Hat
            reason (str): a string when nothing is available to check
        """
        return self._reasons

    @property
    def product(self):
        """ Alias of `release`. Keep backward compatible """
        return self._release

    def __repr__(self):
        if self.is_rhel:
            return "<release: {0}, name: {1}>".format(self.release, self.name)
        return "<release: {0}, name: {1}, reasons: {2}>".format(
                self.release, self.name, self.reasons)



import pytest
import os
from collections import defaultdict

from insights.core import filters
from insights.core.exceptions import SkipComponent
from insights.core import dr
from insights.specs import Specs
from insights.specs.datasources.ls import files_dirs_number


TEST_DIR1 = "/tmp/test_files_number1/"
TEST_DIR2 = "/tmp/test_files_number2"
TEST_DIR3 = "/tmp/test_files_number3"
TEST_DIR4 = "/tmp/test_files_number4"
CREATE_TEST_FILES = "touch " + TEST_DIR1 + "file1" + " " + TEST_DIR1 + "file2"
REMOVE_DIR = "rm -rf " + TEST_DIR1 + " " + TEST_DIR2


def setup_function(func):
    if func is test_module_filters:
        filters.add_filter(Specs.files_dirs_number_filter, [TEST_DIR1, TEST_DIR2, TEST_DIR3])
    if func is test_module_filters_empty:
        filters.add_filter(Specs.files_dirs_number_filter, [])
    if func is test_module_no_target_dir_empty:
        filters.add_filter(Specs.files_dirs_number_filter, [TEST_DIR4])


def teardown_function(func):
    filters._CACHE = {}
    filters.FILTERS = defaultdict(dict)


def test_module_filters():
    broker = dr.Broker()
    os.makedirs(TEST_DIR1)
    os.makedirs(TEST_DIR2)
    os.makedirs(TEST_DIR1 + "test_dir1")
    os.system(CREATE_TEST_FILES)
    result = files_dirs_number(broker)
    os.system(REMOVE_DIR)
    assert result.content == [
        '{"/tmp/test_files_number1/": {"dirs_number": 1, "files_number": 2}, "/tmp/test_files_number2/": {"dirs_number": 0, "files_number": 0}}'
    ]


def test_module_filters_empty():
    broker = dr.Broker()
    with pytest.raises(SkipComponent):
        files_dirs_number(broker)


def test_module_no_target_dir_empty():
    broker = dr.Broker()
    with pytest.raises(SkipComponent):
        files_dirs_number(broker)



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import rpm_v_packages
from insights.parsers.rpm_v_packages import RpmVPackage
from insights.tests import context_wrap


TEST_RPM_V_PACKAGE_1 = """
..?......  c /etc/sudoers
..?......    /usr/bin/sudo
..?......    /usr/bin/sudoreplay
missing     /var/db/sudo/lectured (Permission denied)
error: db5 error(-30973) from dbenv->failchk: BDB0087 DB_RUNRECOVERY: Fatal error, run database recovery
"""

TEST_RPM_V_PACKAGE_2 = """
package sudo is not installed
"""

TEST_RPM_V_PACKAGE_3 = """
error: rpmdb: BDB0113 Thread/process 259/139 failed: BDB1507 Thread died in Berkeley DB library
error: db5 error(-30973) from dbenv->failchk: BDB0087 DB_RUNRECOVERY: Fatal error, run database recovery
error: cannot open Packages index using db5 - (-30973)
"""

TEST_RPM_V_PACKAGE_4 = """
error: rpmdb: BDB0113 Thread/process 259/139 failed: BDB1507 Thread died in Berkeley DB library
error: db5 error(-30973) from dbenv->failchk: BDB0087 DB_RUNRECOVERY: Fatal error, run database recovery
error: cannot open Packages index using db5 - (-30973)

package sudo is not installed
"""

CONTEXT_PATH_1 = "insights_commands/rpm_-V_sudo"


def test_rpm_pkg():
    line = [
        {
            'attributes': '..?......',
            'file': '/etc/sudoers',
            'line': '..?......  c /etc/sudoers',
            'mark': 'c',
        },
        {
            'attributes': '..?......',
            'file': '/usr/bin/sudo',
            'line': '..?......    /usr/bin/sudo',
            'mark': None,
        },
        {
            'attributes': '..?......',
            'file': '/usr/bin/sudoreplay',
            'line': '..?......    /usr/bin/sudoreplay',
            'mark': None,
        },
        {
            'attributes': None,
            'file': None,
            'line': 'missing     /var/db/sudo/lectured (Permission denied)',
            'mark': None,
        },
    ]

    rpm_pkg = RpmVPackage(context_wrap(TEST_RPM_V_PACKAGE_1, CONTEXT_PATH_1))
    for index in range(4):
        assert rpm_pkg.discrepancies[index] == line[index]
    assert rpm_pkg.package_name == 'sudo'


def test_rpm_pkg_empty():
    with pytest.raises(SkipComponent) as exc:
        RpmVPackage(context_wrap([], CONTEXT_PATH_1))
    assert 'Empty result' in str(exc)


def test_rpm_pkg_erros():
    rpm_pkg = RpmVPackage(context_wrap(TEST_RPM_V_PACKAGE_3, CONTEXT_PATH_1))
    assert rpm_pkg.package_name == 'sudo'
    assert rpm_pkg.error_lines is not None
    assert len(rpm_pkg.error_lines) == 3


def test_rpm_pkg_erros_other_lines():
    rpm_pkg = RpmVPackage(context_wrap(TEST_RPM_V_PACKAGE_4, CONTEXT_PATH_1))
    assert rpm_pkg.package_name == 'sudo'
    assert rpm_pkg.discrepancies == []
    assert len(rpm_pkg.error_lines) == 3
    assert "package sudo is not installed" not in rpm_pkg.error_lines


def test_rpm_pkg_not_installed():
    with pytest.raises(SkipComponent) as exc:
        RpmVPackage(context_wrap(TEST_RPM_V_PACKAGE_2, CONTEXT_PATH_1))
    assert 'Package is not installed' in str(exc)


def test_doc_examples():
    env = {"rpm_v_pkg": RpmVPackage(context_wrap(TEST_RPM_V_PACKAGE_1, CONTEXT_PATH_1))}
    failed, total = doctest.testmod(rpm_v_packages, globs=env)
    assert failed == 0



from insights.tests import context_wrap
from insights.parsers import crypto_policies
from insights.parsers.crypto_policies import CryptoPoliciesOpensshserver, CryptoPoliciesConfig, \
    CryptoPoliciesStateCurrent, CryptoPoliciesBind
import doctest


OPENSSHSERVER = """
CRYPTO_POLICY='-oCiphers=aes256-gcm@openssh.com,3des-cbc -oMACs=umac-128-etm@openssh.com'
""".strip()

CONFIG = """
LEGACY
""".strip()

STATECURRENT = """
LEGACY
""".strip()

BIND = """
disable-algorithms "." {
RSAMD5;
DSA;
};
disable-ds-digests "." {
GOST;
};
""".strip()


def test_crypto_policies_doc():
    env = {
        'cp_os': CryptoPoliciesOpensshserver(context_wrap(OPENSSHSERVER)),
        'cp_c': CryptoPoliciesConfig(context_wrap(CONFIG, path="/etc/crypto-policies/config")),
        'cp_sc': CryptoPoliciesStateCurrent(context_wrap(STATECURRENT, path="/etc/crypto-policies/state/current")),
        'cp_bind': CryptoPoliciesBind(context_wrap(BIND)),
    }
    failed, total = doctest.testmod(crypto_policies, globs=env)
    assert failed == 0



"""
Keyword Replacement
===================
"""

import logging
import os

from insights.cleaner.utilities import write_report

logger = logging.getLogger(__name__)


class Keyword(object):
    """
    Class to replace the keyword specified by users to "keyword#"
    """

    def __init__(self, keywords=None):
        # - Keyword replacement information
        self._kw_key = "keyword"
        self._kw_db = dict()  # keyword database
        self._keywords2db(keywords)
        self._obfuscated = set()  # keywords that have been replaced

    def _keywords2db(self, keywords):
        # processes optional keywords to add to be obfuscated
        try:
            if keywords:
                k_count = 0
                for keyword in keywords:
                    keyword = keyword.strip()
                    o_kw = "{0}{1}".format(self._kw_key, k_count)
                    self._kw_db[keyword] = o_kw
                    logger.debug("Added Obfuscated Keyword - %s", o_kw)
                    k_count += 1
                logger.debug("Added All keyword Contents from Customer's configuration")
        except Exception as e:  # pragma: no cover
            logger.warning(e)

    def parse_line(self, line, **kwargs):
        if not line:
            return line
        for k, v in self._kw_db.items():
            if k in line:
                logger.debug("Replacing Keyword - %s > %s", k, v)
                line = line.replace(k, v)
                self._obfuscated.add(k)
        return line

    def mapping(self):
        mapping = []
        for k in self._obfuscated:
            mapping.append({'original': k, 'obfuscated': self._kw_db[k]})
        return mapping

    def generate_report(self, report_dir, archive_name):
        try:
            kw_report_file = os.path.join(report_dir, "%s-keyword.csv" % archive_name)
            logger.info('Creating Keyword Report - %s', kw_report_file)
            lines = ['Replaced Keyword,Original Keyword']
            for k in self._obfuscated:
                lines.append('{0},{1}'.format(k, self._kw_db[k]))
        except Exception as e:  # pragma: no cover
            logger.exception(e)
            raise Exception('CreateReport Error: Error Creating Keyword Report')
        write_report(lines, kw_report_file)
        logger.info('Completed Keyword Report.')



import doctest
import pytest

from insights.tests import context_wrap
from insights.parsers import proc_random_entropy_avail
from insights.core.exceptions import SkipComponent

ENTROPY_CONTENT = """3137"""

WRONG_CONTENT_1 = ""

WRONG_CONTENT_2 = """
abc
def
""".strip()


def test_duplicate_id():
    entropy_obj = proc_random_entropy_avail.RandomEntropyAvail(context_wrap(ENTROPY_CONTENT))
    assert entropy_obj.avail_entropy == 3137


def test_exception():
    with pytest.raises(SkipComponent):
        proc_random_entropy_avail.RandomEntropyAvail(context_wrap(WRONG_CONTENT_1))

    with pytest.raises(SkipComponent):
        proc_random_entropy_avail.RandomEntropyAvail(context_wrap(WRONG_CONTENT_2))


def test_doc():
    random_entropy_obj = proc_random_entropy_avail.RandomEntropyAvail(context_wrap(ENTROPY_CONTENT))
    env = {'random_entropy_obj': random_entropy_obj}
    failed, total = doctest.testmod(proc_random_entropy_avail, globs=env)
    assert failed == 0



"""
This module contains logic for parsing ls output. It attempts to handle
output when selinux is enabled or disabled and also skip "bad" lines.
"""

import re


def parse_path(path):
    """
    Convert possible symbolic link into a source -> target pair.

    Args:
        path (str): The path portion of an ls output line.

    Returns:
        A (path, link) tuple where path is always populated and link is a non
        empty string if the original path is a symoblic link.
    """
    path, _, link = path.partition(" -> ")
    return path, link


def parse_non_selinux(parts):
    """
    Parse part of an ls output line that isn't selinux.

    Args:
        parts (list): A four element list of strings representing the initial
            parts of an ls line after the permission bits. The parts are link
            count, owner, group, and everything else.

    Returns:
        A dict containing links, owner, group, date, and name. If the line
        represented a device, major and minor numbers are included.  Otherwise,
        size is included. If the raw name was a symbolic link, link is
        included.
    """
    links, owner, group, last = parts
    result = {
        "links": int(links),
        "owner": owner,
        "group": group,
    }

    # device numbers only go to 256.
    # If a comma is in the first four characters, the next two elements are
    # major and minor device numbers. Otherwise, the next element is the size.
    if "," in last[:4]:
        major, minor, rest = last.split(None, 2)
        result["major"] = int(major.rstrip(","))
        result["minor"] = int(minor)
    else:
        size, rest = last.split(None, 1)
        result["size"] = int(size)

    # The date part is always 12 characters regardless of content.
    result["date"] = rest[:12]

    # Jump over the date and the following space to get the path part.
    path, link = parse_path(rest[13:])
    result["name"] = path
    if link:
        result["link"] = link

    return result


def parse_selinux(parts):
    """
    Parse part of an ls output line that is selinux.

    Args:
        parts (list): A four element list of strings representing the initial
            parts of an ls line after the permission bits. The parts are owner
            group, selinux info, and the path.

    Returns:
        A dict containing owner, group, se_user, se_role, se_type, se_mls, and
        name. If the raw name was a symbolic link, link is also included.

    """

    owner, group = parts[:2]
    selinux = parts[2].split(":")
    lsel = len(selinux)
    path, link = parse_path(parts[-1])
    result = {
        "owner": owner,
        "group": group,
        "se_user": selinux[0],
        "se_role": selinux[1] if lsel > 1 else None,
        "se_type": selinux[2] if lsel > 2 else None,
        "se_mls": selinux[3] if lsel > 3 else None,
        "name": path,
    }
    if link:
        result["link"] = link
    return result


def parse_rhel8_selinux(parts):
    """
    Parse part of an ls output line that is selinux on RHEL8.

    Args:
        parts (list): A four element list of strings representing the initial
            parts of an ls line after the permission bits. The parts are link
            count, owner, group, and everything else

    Returns:
        A dict containing links, owner, group, se_user, se_role, se_type,
        se_mls, size, date, and name. If the raw name was a symbolic link,
        link is also included.

    """

    links, owner, group, last = parts
    result = {
        "links": int(links),
        "owner": owner,
        "group": group,
    }
    selinux, last = parts[-1].split(None, 1)
    selinux = selinux.split(":")
    lsel = len(selinux)
    if "," in last:
        major, minor, last = last.split(None, 2)
        result['major'] = int(major.rstrip(","))
        result['minor'] = int(minor)
    else:
        size, last = last.split(None, 1)
        result['size'] = int(size)
    date = last[:12]
    path, link = parse_path(last[13:])
    result.update(
        {
            "se_user": selinux[0],
            "se_role": selinux[1] if lsel > 1 else None,
            "se_type": selinux[2] if lsel > 2 else None,
            "se_mls": selinux[3] if lsel > 3 else None,
            "name": path,
            "date": date,
        }
    )
    if link:
        result["link"] = link
    return result


class Directory(dict):
    def __init__(self, name, total, body):
        dirs = []
        ents = {}
        files = []
        specials = []
        for line in body:
            # we can't split(None, 5) here b/c rhel 6/7 selinux lines only have
            # 4 parts before the path, and the path itself could contain
            # spaces. Unfortunately, this means we have to split the line again
            # below
            parts = line.split(None, 4)
            perms = parts[0]
            typ = perms[0]
            entry = {"type": typ, "perms": perms[1:]}
            if parts[1][0].isdigit():
                # We have to split the line again to see if this is a RHEL8
                # selinux stanza. This assumes that the context section will
                # always have at least two pieces separated by ':'.
                # '?' as the whole RHEL8 security context is also acceptable.
                rhel8_selinux_ctx = line.split()[4].strip()
                if ":" in rhel8_selinux_ctx or '?' == rhel8_selinux_ctx:
                    rest = parse_rhel8_selinux(parts[1:])
                else:
                    rest = parse_non_selinux(parts[1:])
            else:
                rest = parse_selinux(parts[1:])

            # Update our entry and put it into the correct buckets
            # based on its type.
            entry.update(rest)
            entry["dir"] = name
            nm = entry["name"]
            ents[nm] = entry
            if typ not in "bcd":
                files.append(nm)
            elif typ == "d":
                dirs.append(nm)
            elif typ in "bc":
                specials.append(nm)

        super(Directory, self).__init__(
            {
                "dirs": dirs,
                "entries": ents,
                "files": files,
                "name": name,
                "specials": specials,
                "total": total,
            }
        )


def parse(lines, root=None):
    """
    Parses a list of lines from ls into dictionaries representing their
    components.

    Args:
        lines (list): A list of lines generated by ls.
        root (str): The directory name to be used for ls output stanzas that
            don't have a name.

    Returns:
        A dictionary representing the ls output. It's keyed by the path
        containing each ls stanza.
    """
    doc = {}
    entries = []
    name = None
    total = None
    for line in lines:
        line = line.strip()
        # Skip empty line and non-exist dir line
        if not line or ': No such file or directory' in line:
            continue
        if line and line[0] == "/" and line[-1] == ":":
            if name is None:
                name = line[:-1]
                if entries:
                    d = Directory(name, total or len(entries), entries)
                    doc[root] = d
                    total = None
                    entries = []
            else:
                d = Directory(name, total or len(entries), entries)
                doc[name or root] = d
                total = None
                entries = []
                name = line[:-1]
            continue
        if line.startswith("total"):
            total = int(line.split(None, 1)[1])
            continue
        entries.append(line)
    name = name or root
    total = total if total is not None else len(entries)
    doc[name] = Directory(name, total, entries)
    return doc


class FilePermissions(object):
    """
    Class for parsing `ls -l` line targeted at concrete file and handling parsed properties.

    It is useful for checking file permissions and owner.

    Attributes:
        type (str): The type indicator (d, c, b, l or -)
        perms_owner (str): Owner permissions, e.g. 'rwx'
        perms_group (str): Group permissions
        perms_other (str): Other permissions
        owner (str): Owner user name
        group (str): Owner group name
        path (str): Full path to file

    Note:
        This class does not support Access Control Lists (ACLs). If that is needed in the future,
        it would be preferable to create another class than extend this one.
        Advanced File Permissions - SUID, SGID and Sticky Bit - are not yet correctly parsed.
    """

    _PERMISSIONS_PATTERN = re.compile(
        r'''
        ^
        (.)([-rwxsStT]{3})([-rwxsStT]{3})([-rwxsStT]{3})   # -rwxrwxrwx
        # -rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config
        # ^^^^^^^^^^
        # Valid characters are -rwxsStT
        #   s == setuid/setgid
        #   S == setuid/setgid without execute permissions
        #   t == sticky bit
        #   T == sticky bit without execute bit

        \S*                      # the character(s) after rwxrwxrwx for ACLs/xattrs
        # -rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config
        #           ^

        \s+\S+\s+                # the number of hardlinks and spaces around
        # -rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config
        #            ^^^

        ([^\s:]+)\s+([^\s:]+)    # owner, spaces, group
        # -rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config
        #               ^^^^^^^^^
        # Username and group name are strings without whitespace \s and without colon :.

        \s+(?:\d+|\d+,\s*\d+)+\s+        # (size or major/minor) and spaces around
        # -rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config
        #                        ^^^^^^
        # brw-rw----. 1 root disk 252, 1 May 16 01:30 /dev/vda1
        #                        ^^^^^^^^

        \S+\s+\S+                # month and day
        # -rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config
        #                              ^^^^^^

        \s+\S+\s+                # time/year and spaces around
        # -rw-------. 1 root root 4308 Apr 22 2009 /etc/ssh/sshd_config
        #                                    ^^^^^^
        # -rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config
        #                                    ^^^^^^^

        (.*)                     # file name or path
        # -rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config
        #                                           ^^^^^^^^^^^^^^^^^^^^
        # -rw-------. 1 root root 4308 Apr 22 15:57 file_name_without_path
        #                                           ^^^^^^^^^^^^^^^^^^^^^^

        $
        ''',
        re.VERBOSE,
    )

    def __init__(self, line):
        """
        Args:
            line (str): A line from `ls -l /concrete/file` execution. Such as:
                        -rw-------. 1 root root 762 Sep 23 002 /etc/ssh/sshd_config
                        -rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config
                        -rw-r--r--. 1 root root 4179 Dec  1  2014 /boot/grub2/grub.cfg
                        brw-rw----. 1 root disk 252, 1 May 16 01:30 /dev/vda1
        Raises:
            ValueError: If line is malformed
        """
        self.line = line
        r = self._PERMISSIONS_PATTERN.search(self.line)
        if r:
            (
                self.type,
                self.perms_owner,
                self.perms_group,
                self.perms_other,
                self.owner,
                self.group,
                self.path,
            ) = r.groups()
            parts = self.line.split()
            if "," in parts[4]:
                self.major = int(parts[4].strip(","))
                self.minor = int(parts[5])
            else:
                self.size = int(parts[4])
        else:
            raise ValueError('Invalid `ls -l` line "{}"'.format(self.line))

    @classmethod
    def from_dict(cls, dirent):
        """
        Create a new FilePermissions object from the given dictionary.  This
        works with the FileListing parser class, which has already done the
        hard work of pulling many of these fields out.  We create an object
        with all the dictionary keys available as properties, and also split
        the ``perms`` string up into owner, group
        """
        # Get a new instance without having to provide a line.
        self = cls.__new__(cls)
        # Check that we have at least as much data as the __init__ requires
        for k in ['type', 'perms', 'owner', 'group', 'name', 'dir']:
            if k not in dirent:
                raise ValueError("Need required key '{k}'".format(k=k))
        # Copy all values across
        for key, val in dirent.items():
            setattr(self, key, val)
        # Fill in names that might not have been passed from the ls parser
        self.path = self.name
        # Create perms parts
        self.perms_owner = self.perms[0:3]
        self.perms_group = self.perms[3:6]
        self.perms_other = self.perms[6:9]
        return self

    def owned_by(self, owner, also_check_group=False):
        """
        Checks if the specified user or user and group own the file.

        Args:
            owner (str): the user (or group) name for which we ask about ownership
            also_check_group (bool): if set to True, both user owner and group owner checked
                                if set to False, only user owner checked

        Returns:
            bool: True if owner of the file is the specified owner
        """
        if also_check_group:
            return self.owner == owner and self.group == owner
        else:
            return self.owner == owner

    def owner_can_read(self):
        """
        Checks if owner can read the file. Write and execute bits are not evaluated.

        Returns:
            bool: True if owner can read the file.
        """
        return 'r' in self.perms_owner

    def group_can_read(self):
        """
        Checks if group can read the file. Write and execute bits are not evaluated.

        Returns:
            bool: True if group can read the file.
        """
        return 'r' in self.perms_group

    def others_can_read(self):
        """
        Checks if 'others' can read the file. Write and execute bits are not evaluated. ('others' in
        the sense of unix permissions that know about user, group, others.)

        Returns:
            bool: True if 'others' can read the file.
        """
        return 'r' in self.perms_other

    def owner_can_only_read(self):
        """
        Checks if owner has read-only permissions for the file.
        Therefore, write and execute bits for owner must be unset and read bit must be set.

        Returns:
            bool: True if owner can only read the file.
        """
        return 'r--' == self.perms_owner

    def group_can_only_read(self):
        """
        Checks if group has read-only permissions for the file.
        Therefore, write and execute bits for group must be unset and read bit must be set.

        Returns:
            bool: True if group can only read the file.
        """
        return 'r--' == self.perms_group

    def others_can_only_read(self):
        """
        Checks if 'others' has read-only permissions for the file.
        Therefore, write and execute bits for 'others' must be unset and read bit must be set.
        ('others' in the sense of unix permissions that know about user, group, others.)

        Returns:
            bool: True if 'others' can only read the file.
        """
        return 'r--' == self.perms_other

    def owner_can_write(self):
        """
        Checks if owner can write the file. Read and execute bits are not evaluated.

        Returns:
            bool: True if owner can write the file.
        """
        return 'w' in self.perms_owner

    def group_can_write(self):
        """
        Checks if group can write the file. Read and execute bits are not evaluated.

        Returns:
            bool: True if group can write the file.
        """
        return 'w' in self.perms_group

    def others_can_write(self):
        """
        Checks if 'others' can write the file. Read and execute bits are not evaluated. ('others' in
        the sense of unix permissions that know about user, group, others.)

        Returns:
            bool: True if 'others' can write the file.
        """
        return 'w' in self.perms_other

    def only_root_can_read(self, root_group_can_read=True):
        """
        Checks if only root is allowed to read the file (and anyone else is
        forbidden from reading). Write and execute bits are not checked. The
        read bits for root user/group are not checked because root can
        read/write anything regardless of the read/write permissions.

        When called with ``root_group_can_read`` = ``True``:

        * owner must be root
        * and 'others' permissions must not contain read
        * and if group owner is not root, the 'group' permissions must not
          contain read

        Valid cases::

            rwxrwxrwx    owner   ownergroup
            -------------------------------
            ???-??-??    root    nonroot
            ??????-??    root    root
            r--r-----    root    root
            r--------    root    nonroot
            rwxrwx---    root    root
            rwxrwx-wx    root    root

        Specifically, these cases are NOT valid because the owner can chmod
        permissions and grant themselves permissions without root's
        knowledge::

            rwxrwxrwx    owner   ownergroup
            -------------------------------
            -??-??-??    nonroot nonroot
            -??r??-??    nonroot root
            ---------    nonroot nonroot

        When called with ``root_group_can_read`` = ``False``:

        * owner must be root
        * and 'group' and 'others' permissions must not contain read

        Valid cases::

            rwxrwxrwx    owner   ownergroup
            -------------------------------
            ???-??-??    root    ?
            r--------    root    root
            r--------    root    nonroot
            rwx-wx---    root    root
            rwx-wx---    root    nonroot
            rwx-wxrwx    root    nonroot

        Specifically, these cases are NOT valid because the owner can chmod
        permissions and grant themselves permissions without root's
        knowledge::

            rwxrwxrwx    owner   ownergroup
            -------------------------------
            -??-??-??    nonroot nonroot
            ---------    nonroot nonroot

        Args:
            root_group_can_read (bool): if set to True, this tests whether the
            'root' group can also read the file.

        Returns:
            bool: True if only root user (or optionally root group) can read
            the file.
        """

        requirements = True  # The final answer is progressively assembled in this variable.
        requirements &= self.owner == 'root'
        requirements &= not self.others_can_read()
        if root_group_can_read:
            if self.group != 'root':
                # if group is not root, group must not be able to read
                requirements &= not self.group_can_read()
        else:  # root_group_can_read == False
            requirements &= not self.group_can_read()
        return requirements

    def only_root_can_write(self, root_group_can_write=True):
        """
        Checks if only root is allowed to write the file (and anyone else is
        barred from writing). Read and execute bits are not checked. The
        write bits for root user/group are not checked because root can
        read/write anything regardless of the read/write permissions.

        When called with ``root_group_can_write`` = ``True``:

        * owner must be root
        * and 'others' permissions must not contain write
        * and if group owner is not root, the 'group' permissions must not contain write

        Valid cases::

            rwxrwxrwx    owner   ownergroup
            -------------------------------
            ????-??-?    root    nonroot
            ???????-?    root    root
            -w--w----    root    root
            -w-------    root    root
            rwxrwx---    root    root
            rwxrwxr-x    root    root

        Specifically, these cases are NOT valid because the owner can chmod
        permissions and grant themselves permissions without root's
        knowledge::

            rwxrwxrwx    owner   ownergroup
            -------------------------------
            ?-??-??-?    nonroot nonroot
            ?-??w??-?    nonroot root
            ---------    nonroot nonroot

        When called with ``root_group_can_write`` = ``False``:

        * owner must be root
        * and 'group' and 'others' permissions must not contain write

        Valid cases::

            rwxrwxrwx    owner   ownergroup
            -------------------------------
            ????-??-?    root    ?
            -w-------    root    root
            -w-------    root    nonroot
            rwxr-x---    root    root
            rwxr-x---    root    nonroot
            rwxr-xr-x    root    nonroot

        Specifically, these cases are NOT valid because the owner can chmod
        permissions and grant themselves permissions without root's
        knowledge::

            rwxrwxrwx    owner   ownergroup
            -------------------------------
            ?-??-??-?    nonroot nonroot
            ---------    nonroot nonroot

        Args:
            root_group_can_write (bool): if set to True, this tests whether
            'root' group can also write to the file.

        Returns:
            bool: True if only root user (or optionally root group) can write
            the file.
        """

        requirements = True  # The final answer is progressively assembled in this variable.
        requirements &= self.owner == 'root'
        requirements &= not self.others_can_write()
        if root_group_can_write:
            if self.group != 'root':
                # if group is not root, group must not be able to write
                requirements &= not self.group_can_write()
        else:  # root_group_can_write == False
            requirements &= not self.group_can_write()
        return requirements

    def all_zero(self):
        """
        Checks that all permissions are zero ('---------' in ls -l) - nobody but root can read,
        write, exec.

        Returns:
            bool: True if all permissions are zero ('---------')
        """
        _PERM_NOTHING = '---'
        return all(
            (
                self.perms_owner == _PERM_NOTHING,
                self.perms_group == _PERM_NOTHING,
                self.perms_other == _PERM_NOTHING,
            )
        )

    def __repr__(self):
        return 'FilePermissions(' + self.path + ')'



"""
System class files under ``/sys/class``
=======================================

This module contains the following parsers:

TtyConsoleActive - file ``/sys/class/tty/console/active``
---------------------------------------------------------
"""
from insights.core import Parser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.tty_console_active)
class TtyConsoleActive(Parser):
    """
    Parser for the `/sys/class/tty/console/active` file.

    Sample Content::

        tty0 ttyS0

    Raises:
        SkipComponent: When content has more than one lines.

    Attributes:
        devices(list): a list of kernel console devices.

    Examples::

        >>> tty_console_active.devices
        ['tty0', 'ttyS0']
    """

    def parse_content(self, content):
        if len(content) > 1:
            raise SkipComponent("This should be an one line file")
        if not content:
            raise SkipComponent("Empty content.")
        else:
            self.devices = content[0].split()



from insights.parsers.jboss_version import JbossVersion, JbossRuntimeVersions
from insights.tests import context_wrap
import doctest
from insights.parsers import jboss_version

JBOSS_6 = """
Red Hat JBoss Enterprise Application Platform - Version 6.4.3.GA
""".strip()

JBOSS_7 = """
Red Hat JBoss Enterprise Application Platform - Version 7.1.0.Beta1
""".strip()


def test_jboss6():
    release6 = JbossVersion(context_wrap(JBOSS_6, path='/home/test/jboss/jboss-eap-6.4/version.txt'))
    assert release6.file_path == '/home/test/jboss/jboss-eap-6.4/version.txt'
    assert release6.raw == JBOSS_6
    assert release6.major == 6
    assert release6.minor == 4
    assert release6.version == "6.4.3"
    assert release6.code_name == "GA"
    assert release6.release == 3


def test_jboss7():
    release7 = JbossVersion(context_wrap(JBOSS_7, path='/home/test/jboss/jboss-eap-7.1/version.txt'))
    assert release7.file_path == '/home/test/jboss/jboss-eap-7.1/version.txt'
    assert release7.raw == JBOSS_7
    assert release7.major == 7
    assert release7.minor == 1
    assert release7.version == "7.1.0"
    assert release7.code_name == "Beta1"
    assert release7.release == 0


SINGLE_JBOSS = """
{"/opt/jboss-datagrid-7.3.0-server": "Red Hat Data Grid - Version 7.3.0"}
"""

MULTI_JBOSS = """
{"/opt/jboss": "Red Hat JBoss Enterprise Application Platform - Version 7.3.0.GA\n", "/opt/jboss-sss": "Red Hat Single Sign-On - Version 7.4.10.GA\n", "/opt/jboss-keycloak": "Keycloak - Version 3.4.3.Final\n"}
"""
WEB_SERVER_RUNTIME = """
{"/opt/jboss-web-server": "Red Hat JBoss Web Server - Version 5.6 GA\n\nJava Components:\nApache Tomcat 9.0.50-3.redhat_00004.1.el7jws\nTomcat Vault 1.1.8-4.Final_redhat_00004.1.el7jws\nJBoss mod_cluster 1.4.3-2.Final_redhat_00002.1.el7jws\n\nNative Components:\nApache Tomcat Native 1.2.30-3.redhat_3.el7jws\\nApache Portable Runtime (APR) 1.6.3-107.jbcs.el7\nOpenSSL 1.1.1g-8.jbcs.el7\n"}
"""


def test_JbossRuntimeVersions():
    jboss_runtimes = JbossRuntimeVersions(context_wrap(SINGLE_JBOSS))
    jboss_v = jboss_runtimes[0]
    assert jboss_v.file_path == "/opt/jboss-datagrid-7.3.0-server"
    assert jboss_v.product == "Red Hat Data Grid"
    assert jboss_v.version == "7.3.0"
    assert jboss_v.major == 7
    assert jboss_v.minor == 3
    assert jboss_v.release == 0
    jboss_runtimes = JbossRuntimeVersions(context_wrap(MULTI_JBOSS))
    for j in jboss_runtimes:
        assert 'Red Hat' in j.product or 'Keycloak' in j.product
    jboss_v = jboss_runtimes[0]
    assert jboss_v.file_path == "/opt/jboss"
    assert jboss_v.version == "7.3.0"
    assert jboss_v.code_name == "GA"
    jboss_v = jboss_runtimes[2]
    assert jboss_v.file_path == "/opt/jboss-keycloak"
    assert jboss_v.version == "3.4.3"
    assert jboss_v.code_name == "Final"
    jboss_runtimes = JbossRuntimeVersions(context_wrap(WEB_SERVER_RUNTIME))
    jboss_v = jboss_runtimes[0]
    assert jboss_v.product == "Red Hat JBoss Web Server"
    assert jboss_v.file_path == "/opt/jboss-web-server"
    assert jboss_v.code_name == "GA"
    assert jboss_v.version == "5.6"
    assert jboss_v.major == 5
    assert jboss_v.minor == 6
    assert jboss_v.release == 0


def test_jboss_version_doc_examples():
    env = {
        'jboss_version': JbossVersion(context_wrap(JBOSS_6,
                                                   path='/home/test/jboss/jboss-eap-6.4/version.txt')),
        'all_jboss_versions': JbossRuntimeVersions(context_wrap(SINGLE_JBOSS,
                                                                path='/home/test/jboss/jboss_versions'))
    }
    failed, total = doctest.testmod(jboss_version, globs=env)
    assert failed == 0



#!/usr/bin/env python
"""
This module uses the inspec module to introspec information about
all of the datasources and dynamically create a ``.rst`` document to be
included in the sphinx build process

The output file is a listing of all default specs defined in
:py:mod:`insights.specs.default`.

It can be called from the commandline to manually generate the file or
it may also be called inside the Sphinx *conf.py* `setup` function to
generate the documentation automatically.

Command Line Use
----------------

This assumes that you have a working installation of insights-core
and python environment for insights-core development.

To run from the command line simply execute this utility and provide
the name of the output file::

    $ python -m insights.util.specs_catalog.main output_filename.rst

"""
import argparse
from insights.util.autology.datasources import DefaultSpecs

TEMPLATE = """
Datasource Catalog
==================

This catalog of specs provides the definitions for each of the specs the Insights Client
can collect. The format for this file is ``name = collection_type(collection_arguments)``.
A detailed description of the collection_types can be found in :ref:`datasources-ref`.
A summary of the collection_types includes:

* *simple_file* - collects the contents of a file
* *simple_command* - collects the output from a command
* *glob_file* - collects contents of all files matching the glob pattern
* *first_file* - collects the contents of the first file detected in a list
  of files
* *listdir* - collects the output of the ``ls`` command for the file/glob argument
* *foreach_execute* - collects the output of the command for each ``provider`` argument
* *foreach_collect* - collects the contents of the path created by replacing
  each element in the provider into the path
* *first_of* - collects the contents of datasource that returns data
* *command_with_args* - collects the output of the command with each ``provider`` argument
* *head* - collects the contents of the first item in a list

Some datasources are implemented as functions and each links to the details provided in the
function specific documentation.  Generally functions are used as a ``provider`` to other
datasources to, for instance, get a list of running processes of a particular program.
Functions implemented as custom datasources can be found in :ref:`custom-datasources`.

Python code that implements these datasources is located in the module
:py:mod:`insights.specs.default`.  The datasources each have a unique name in the
class :py:class:`insights.specs.default.DefaultSpecs`.

Datasources
-----------

Functions
^^^^^^^^^

.. hlist::
""".strip()


def blank_line(fh, number=1):
    for l in range(number):
        fh.write("\n")


def main(filename):
    defaultspecs = DefaultSpecs()
    with open(filename, "w") as fh:
        fh.write(TEMPLATE)
        blank_line(fh, 2)
        functions = [v for v in defaultspecs.values() if v.is_function]
        for v in functions:
            try:
                fh.write('    * :py:func:`{fxn_name}() <insights.specs.default.DefaultSpecs.{fxn_name}>`\n'.format(fxn_name=v['fxn_name']))
            except Exception:
                print('Error with function spec: {name}'.format(name=v['fxn_name']))

        fh.write('\n\nGeneral Datasources\n^^^^^^^^^^^^^^^^^^^\n\n::\n\n')

        for k, v in defaultspecs.items():
            try:
                if not v.is_function:
                    line = str(v)
                    line = line.replace('\n', '\\n')
                    fh.write('    {spec}\n'.format(spec=line))
            except Exception:
                print('Error with spec: {name}'.format(name=k))


def parse_args():
    p = argparse.ArgumentParser()
    p.add_argument("out_filename", help="Name of the output file to write the specs catalog")
    return p.parse_args()


if __name__ == "__main__":
    args = parse_args()
    main(args.out_filename)



from insights.parsers.openshift_kube_log import ControllerManagerLog
from insights.parsers.openshift_kube_log import ApiServerLog
from insights.tests import context_wrap


OPENSHIFT_KUBE_LOG = """
2023-12-18T06:13:36.831359143+00:00 stderr F I1218 06:13:36.831274      16 aggregator.go:115] Building initial OpenAPI spec
2023-12-18T06:13:36.846052152+00:00 stderr F I1218 06:13:36.845970      16 aggregator.go:118] Finished initial OpenAPI spec generation after 14.671324ms
2024-12-18T06:13:37.031688793+00:00 stderr F I1218 06:13:37.031622      16 genericapiserver.go:535] "[graceful-termination] using HTTP Server shutdown timeout" ShutdownTimeout="2s"
2024-12-18T06:13:37.031809098+00:00 stderr F I1218 06:13:37.031766      16 dynamic_cafile_content.go:157] "Starting controller" name="client-ca-bundle::/etc/kubernetes/static-pod-certs/configmaps/client-ca/ca-bundle.crt"
2023-12-18T06:13:37.031901208+00:00 stderr F I1218 06:13:37.031766      16 dynamic_cafile_content.go:157] "Starting controller" name="request-header::/etc/kubernetes/static-pod-certs/configmaps/aggregator-client-ca/ca-bundle.crt"
"""


def test_openshift_kube_controller_manager_log():
    log = ControllerManagerLog(context_wrap(OPENSHIFT_KUBE_LOG))
    assert "HTTP Server shutdown timeout" in log


def test_openshift_kube_api_server_log():
    log = ApiServerLog(context_wrap(OPENSHIFT_KUBE_LOG))
    assert "Building initial OpenAPI spec" in log



import doctest
import pytest

from insights.core.exceptions import ParseException
from insights.parsers import kubepods_cpu_quota
from insights.parsers.kubepods_cpu_quota import KubepodsCpuQuota
from insights.tests import context_wrap

KUBEPODS_CPU_QUOTA_1 = """
-1
""".strip()

KUBEPODS_CPU_QUOTA_2 = """
50000
""".strip()

KUBEPODS_CPU_QUOTA_INVALID = """
invalid
-1
""".strip()


def test_kubepods_cpu_quota():
    cpu_quota = kubepods_cpu_quota.KubepodsCpuQuota(context_wrap(KUBEPODS_CPU_QUOTA_1))
    assert cpu_quota.cpu_quota == -1


def test_kubepods_cpu_quota_2():
    cpu_quota = kubepods_cpu_quota.KubepodsCpuQuota(context_wrap(KUBEPODS_CPU_QUOTA_2))
    assert cpu_quota.cpu_quota == 50000


def test_invalid():
    with pytest.raises(ParseException) as e:
        kubepods_cpu_quota.KubepodsCpuQuota(context_wrap(KUBEPODS_CPU_QUOTA_INVALID))
    assert "invalid" in str(e)


def test_akubepods_cpu_quota_doc_examples():
    env = {
        'kubepods_cpu_quota': KubepodsCpuQuota(
            context_wrap(KUBEPODS_CPU_QUOTA_1)),
    }
    failed, total = doctest.testmod(kubepods_cpu_quota, globs=env)
    assert failed == 0



"""
Utilities for spec Cleaner
==========================
"""

import logging
import json
import os

logger = logging.getLogger(__name__)


def write_report(report, report_file, mode=0o644):
    # Get the current umask
    umask = os.umask(0o022)
    # Reset the umask
    os.umask(umask)
    try:
        with open(report_file, 'w') as fp:
            if isinstance(report, dict):
                json.dump(report, fp)
            elif isinstance(report, list):
                for line in report:
                    fp.write("{0}\n".format(line))
        # Change the file mode per the current umask
        os.chmod(report_file, mode & ~umask)
    except (IOError, OSError) as e:  # pragma: no cover
        logger.error('Could not write to %s: %s', report_file, str(e))



try:
    from unittest.mock import patch
except Exception:
    from mock import patch

from copy import deepcopy

from insights.client.connection import InsightsConnection
from insights.util.hostname import determine_hostname

from insights.client.config import InsightsConfig


class MockSession(object):
    def __init__(self):
        self.status_code = None
        self.text = None
        self.content = '{"display_name": "test"}'
        self.headers = dict()


def mock_init_session(obj):
    return MockSession()


CFACTS = {
    'hostname': determine_hostname(),
    'ip': ['10.0.0.1', '192.168.0.1'],
}


@patch('insights.client.connection.InsightsConnection._init_session', mock_init_session)
@patch('insights.client.connection.InsightsConnection.post', return_value=MockSession())
@patch('insights.client.connection.InsightsUploadConf.get_rm_conf', return_value={})
def test_connection_clean_facts_no_clean_all(rm_conf, post):
    # No cleaning per client configuration
    config = InsightsConfig(obfuscate=False)
    conn = InsightsConnection(config)
    tmp = deepcopy(CFACTS)
    ret = conn._clean_facts(tmp)
    # Nothing changed
    assert ret == CFACTS


@patch('insights.client.connection.InsightsConnection._init_session', mock_init_session)
@patch('insights.client.connection.InsightsConnection.post', return_value=MockSession())
@patch('insights.client.connection.InsightsUploadConf.get_rm_conf', return_value={})
def test_connection_clean_facts_clean_ip_only(rm_conf, post):
    # No hostname cleaning per client configuration
    config = InsightsConfig(obfuscate=True, obfuscate_hostname=False)
    conn = InsightsConnection(config)
    tmp = deepcopy(CFACTS)
    ret = conn._clean_facts(tmp)
    # Hostname is not changed
    assert ret['hostname'] == CFACTS['hostname']
    # IPs are changed
    assert ret['ip'] != CFACTS['ip']


@patch('insights.client.connection.InsightsConnection._init_session', mock_init_session)
@patch('insights.client.connection.InsightsConnection.post', return_value=MockSession())
@patch('insights.client.connection.InsightsUploadConf.get_rm_conf', return_value={})
def test_connection_clean_facts_clean_all(rm_conf, post):
    config = InsightsConfig(obfuscate=True, obfuscate_hostname=True)
    conn = InsightsConnection(config)
    tmp = deepcopy(CFACTS)
    ret = conn._clean_facts(tmp)
    # Hostname is changed
    assert ret['hostname'] != CFACTS['hostname']
    # IPs are changed
    assert ret['ip'] != CFACTS['ip']



import doctest
import pytest

from insights.core.exceptions import SkipComponent, ParseException
from insights.parsers import ilab
from insights.parsers.ilab import IlabModuleList, IlabConfigShow
from insights.tests import context_wrap

ILAB_MODULE_LIST = """
+-----------------------------------+---------------------+---------+
| Model Name                        | Last Modified       | Size    |
+-----------------------------------+---------------------+---------+
| models/prometheus-8x7b-v2-0       | 2024-08-09 13:28:50 |  87.0 GB|
| models/mixtral-8x7b-instruct-v0-1 | 2024-08-09 13:28:24 |  87.0 GB|
| models/granite-7b-redhat-lab      | 2024-08-09 14:28:40 |  12.6 GB|
| models/granite-7b-starter         | 2024-08-09 14:40:35 |  12.6 GB|
+-----------------------------------+---------------------+---------+
""".strip()

ILAB_MODULE_LIST_ISSUE_FORMAT = """
+-----------------------------------+---------------------+
| Model Name                        | Last Modified       |
+-----------------------------------+---------------------+
| models/mixtral-8x7b-instruct-v0-1 | 2024-08-09 13:28:24 |
| models/granite-7b-redhat-lab      | 2024-08-09 14:28:40 |
| models/granite-7b-starter         | 2024-08-09 14:40:35 |
+-----------------------------------+---------------------+
""".strip()

ILAB_MODULE_LIST_ERROR = """
Usage: ilab model [OPTIONS] COMMAND [ARGS]...
Try 'ilab model --help' for help.

Error: `/root/.config/instructlab/config.yaml` does not exist or is not a readable file.
Please run `ilab config init` or point to a valid configuration file using `--config=<path>`.
""".strip()

ILAB_MODULE_LIST_EMPTY = ""

ILAB_CONFIG_SHOW = """
time="2025-04-15T08:23:44Z" level=warning msg="The input device is not a TTY. The --tty and --interactive flags might not work properly"
# Chat configuration section.
chat:
  # Predefined setting or environment that influences the behavior and responses of
  # the chat assistant. Each context is associated with a specific prompt that
  # guides the assistant on how to respond to user inputs. Available contexts:
  context: default
  logs_dir: /root/.local/share/instructlab/chatlogs
  max_tokens:
  model: /root/.cache/instructlab/models/granite-8b-lab-v1
evaluate:
  base_branch:
  base_model: /root/.cache/instructlab/models/granite-8b-starter-v1
  mmlu:
    batch_size: auto
    few_shots: 5
serve:
  backend: vllm
  vllm:
    gpus:
    llm_family: ''
    vllm_args:
    - --tensor-parallel-size
    - '1'
""".strip()

ILAB_CONFIG_SHOW_ERROR = ILAB_MODULE_LIST_ERROR
ILAB_CONFIG_SHOW_EMPTY = ILAB_MODULE_LIST_EMPTY


def test_ilab_model_list():
    ilab_module_list_info = IlabModuleList(context_wrap(ILAB_MODULE_LIST))
    assert len(ilab_module_list_info) == 4
    assert ilab_module_list_info[0]["model_name"] == "models/prometheus-8x7b-v2-0"
    assert ilab_module_list_info[1]["last_modified"] == "2024-08-09 13:28:24"
    assert ilab_module_list_info[2]["size"] == "12.6 GB"
    assert ilab_module_list_info.models == ["models/prometheus-8x7b-v2-0", "models/mixtral-8x7b-instruct-v0-1", "models/granite-7b-redhat-lab", "models/granite-7b-starter"]

    with pytest.raises(SkipComponent):
        IlabModuleList(context_wrap(ILAB_MODULE_LIST_EMPTY))

    with pytest.raises(ParseException):
        IlabModuleList(context_wrap(ILAB_MODULE_LIST_ERROR))

    with pytest.raises(ParseException):
        IlabModuleList(context_wrap(ILAB_MODULE_LIST_ISSUE_FORMAT))


def test_ilab_config_show():
    ilab_config_show_info = IlabConfigShow(context_wrap(ILAB_CONFIG_SHOW))
    assert ilab_config_show_info['chat']['logs_dir'] == '/root/.local/share/instructlab/chatlogs'
    assert ilab_config_show_info['serve']['vllm']['vllm_args'] == ['--tensor-parallel-size', '1']

    with pytest.raises(ParseException):
        IlabConfigShow(context_wrap(ILAB_CONFIG_SHOW_ERROR))

    with pytest.raises(SkipComponent):
        IlabModuleList(context_wrap(ILAB_CONFIG_SHOW_EMPTY))


def test_ilab_doc_examples():
    env = {
        'ilab_model_list_info': IlabModuleList(context_wrap(ILAB_MODULE_LIST)),
        'ilab_config_show_info': IlabConfigShow(context_wrap(ILAB_CONFIG_SHOW))
    }
    failed, total = doctest.testmod(ilab, globs=env)
    assert failed == 0



"""
smartctl - parser for smartctl commands
=======================================

Classes to parse the output of smartctl commands:


SMARTctl - /usr/sbin/smartctl -a {devices}
------------------------------------------
SmartctlHealth - /usr/sbin/smartctl -H {devices} -j
---------------------------------------------------
"""
import re

from insights.core import CommandParser, JSONParser
from insights.core.exceptions import ParseException
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.smartctl)
class SMARTctl(CommandParser):
    """
    Parser for output of ``smartctl -a`` for each drive in system.

    This stores the information from the output of `smartctl` in the
    following properties:

     * ``device`` - the name of the device after /dev/ - e.g. sda
     * ``information`` - the -i info (vendor, product, etc)
     * ``health`` - overall health assessment (-H)
     * ``values`` - the SMART values (-c) - SMART config on drive firmware
     * ``attributes`` - the SMART attributes (-A) - run time data

    For legacy access, these are also available as values in the ``info``
    dictionary property, keyed to their name (i.e. info['device'])

    Each object contains a different device; the shared information for this
    parser in Insights will be one or more devices, so see the example below
    for how to iterate through the available SMARTctl information for each
    device.

    Sample (abbreviated) output::

        smartctl 6.2 2013-07-26 r3841 [x86_64-linux-3.10.0-267.el7.x86_64] (local build)
        Copyright (C) 2002-13, Bruce Allen, Christian Franke, www.smartmontools.org

        === START OF INFORMATION SECTION ===
        Device Model:     ST500LM021-1KJ152
        Serial Number:    W620AT02
        LU WWN Device Id: 5 000c50 07817bb36
        ...

        === START OF READ SMART DATA SECTION ===
        SMART overall-health self-assessment test result: PASSED

        General SMART Values:
        Offline data collection status:  (0x00) Offline data collection activity
                            was never started.
                            Auto Offline Data Collection: Disabled.
        ...

        SMART Attributes Data Structure revision number: 10
        Vendor Specific SMART Attributes with Thresholds:
        ID# ATTRIBUTE_NAME          FLAG     VALUE WORST THRESH TYPE      UPDATED  WHEN_FAILED RAW_VALUE
          1 Raw_Read_Error_Rate     0x000f   118   099   034    Pre-fail  Always       -       179599704
          3 Spin_Up_Time            0x0003   098   098   000    Pre-fail  Always       -       0
          4 Start_Stop_Count        0x0032   100   100   020    Old_age   Always       -       546
          5 Reallocated_Sector_Ct   0x0033   100   100   036    Pre-fail  Always       -       0
        ...

    Examples:
        >>> type(smartctl_all)
        <class 'insights.parsers.smartctl.SMARTctl'>
        >>> smartctl_all.device
        '/dev/sdc'
        >>> smartctl_all.information['Vendor']
        'NETAPP'

    """

    _INFO_LINE_STR = r'(?P<key>\w+(?:\s\w+)*):\s+' + \
        r'(?P<value>\S.*?)\s*$'
    _INFO_LINE_RE = re.compile(_INFO_LINE_STR)
    _VALUE_LINE_STR = r'(?P<key>\w[A-Za-z _.-]+):\s+' + \
        r'\(\s*(?P<value>\S.*?)\)'
    _VALUE_LINE_RE = re.compile(_VALUE_LINE_STR)
    _ATTR_LINE_STR = r'^\s*(?P<id>\d+)\s(?P<name>\w+)\s+' + \
        r'(?P<flag>0x[0-9a-fA-F]{4})\s+(?P<value>\d{3})\s+' + \
        r'(?P<worst>\d{3})\s+(?P<threshold>\d{3})\s+' + \
        r'(?P<type>[A-Za-z_-]+)\s+(?P<updated>[A-Za-z_-]+)\s+' + \
        r'(?P<when_failed>\S+)\s+(?P<raw_value>\S.*)$'
    _ATTR_LINE_RE = re.compile(_ATTR_LINE_STR)

    def __init__(self, context):
        filename_re = re.compile(r'smartctl_-a_\.dev\.(?P<device>\w+)$')
        match = filename_re.search(context.path)
        if match:
            self.device = '/dev/' + match.group('device')
        else:
            raise ParseException('Cannot parse device name from path {p}'.format(p=context.path))
        super(SMARTctl, self).__init__(context)

    def parse_content(self, content):
        self.information = {}
        self.health = 'not parsed'
        self.values = {}
        self.attributes = {}
        # hack for persistent line storage in parse_content context -
        # otherwise it gets treated as a local variable within the sub-
        # functions
        self.full_line = ''

        # Parsing using a state machine, sorry.  We use a state variable, and
        # functions to parse lines in each of the different states.  The
        # function returns the state as a result of reading that line, and we
        # look up the parse function out of an array based on the parse state.
        PARSE_FORMATTED_INFO = 0
        PARSE_FREEFORM_INFO = 1
        PARSE_ATTRIBUTE_INFO = 2
        PARSE_COMPLETE = 3
        parse_state = PARSE_FORMATTED_INFO

        # Information section:
        def parse_information(line):
            # Exit parsing information section if we go into the next section
            if line.startswith('=== START OF READ SMART DATA SECTION ==='):
                return PARSE_FREEFORM_INFO
            match = self._INFO_LINE_RE.search(line)
            if match:
                self.information[match.group('key')] = match.group('value')
            else:
                # Translate some of the less structured information
                if line == 'Device does not support SMART':
                    self.information['SMART support is'] = 'Not supported'
                elif line == 'Device supports SMART and is Enabled':
                    self.information['SMART support is'] = 'Enabled'
                elif line == 'Error Counter logging not supported':
                    self.information['Error Counter logging'] = \
                        'Not supported'
                elif line == 'Device does not support Self Test logging':
                    self.information['Self Test logging'] = 'Not supported'
                elif line == 'Temperature Warning Disabled or Not Supported':
                    self.information['Temperature Warning'] = \
                        'Disabled or Not Supported'
            return PARSE_FORMATTED_INFO

        # Values section:
        def parse_values(line):
            if line.startswith('Vendor Specific SMART Attributes with Thres'):
                return PARSE_ATTRIBUTE_INFO
            if line.startswith('SMART overall-health self-assessment test r'):
                self.health = ''.join((line.split(': '))[1:])
                return PARSE_FREEFORM_INFO
            # Values section begins with this - ignore:
            if line.startswith('General SMART Values:'):
                return PARSE_FREEFORM_INFO

            # Lines starting with a space are continuations of the commentary
            # on the previous setting - ignore
            if len(line) == 0 or line[0] == ' ' or line[0] == "\t":
                return PARSE_FREEFORM_INFO
            # Otherwise, join this line to the full line
            if self.full_line:
                self.full_line += ' '
            self.full_line += line.strip()

            match = self._VALUE_LINE_RE.search(self.full_line)
            if match:
                # Handle the recommended polling time lines, which are joined
                # with the previous line and values are in minutes.
                (key, value) = match.group('key', 'value')
                self.values[key] = value
                self.full_line = ''
            elif self.full_line.startswith('SMART Attributes Data Structure revision number: '):
                (key, value) = self.full_line.split(': ')
                self.values[key] = value
                self.full_line = ''
            return PARSE_FREEFORM_INFO

        # Attributes sections
        def parse_attributes(line):
            if line.startswith('SMART Error Log Version:'):
                return PARSE_COMPLETE
            if len(line) == 0:
                return PARSE_ATTRIBUTE_INFO
            match = self._ATTR_LINE_RE.match(line)
            if match:
                name = match.group('name')
                self.attributes[name] = match.groupdict()
            return PARSE_ATTRIBUTE_INFO

        parse_for_state = [
            parse_information,
            parse_values,
            parse_attributes,
        ]

        for line in content:
            parse_state = parse_for_state[parse_state](line)
            if parse_state == PARSE_COMPLETE:
                break

        # Delete temporary full line storage
        del self.full_line


@parser(Specs.smartctl_health)
class SmartctlHealth(JSONParser):
    """
    Parse the output of command "smartctl -H -d scsi {devices}".

    Sample input::

        {
          "json_format_version": [
            1,
            0
          ],
          "smartctl": {
            "version": [
              7,
              2
            ],
            "svn_revision": "5155",
            "platform_info": "x86_64-linux-5.14.0-503.11.1.el9_5.x86_64",
            "build_info": "(local build)",
            "argv": [
              "smartctl",
              "-H",
              "-d",
              "scsi",
              "/dev/sdb",
              "-j"
            ],
            "exit_status": 0
          },
          "device": {
            "name": "/dev/sdb",
            "info_name": "/dev/sdb",
            "type": "scsi",
            "protocol": "SCSI"
          },
          "smart_status": {
            "passed": true
          }
        }

    Examples:
        >>> type(smartctl_health)
        <class 'insights.parsers.smartctl.SmartctlHealth'>
        >>> smartctl_health.data['smart_status']["passed"]
        True

    """
    pass



import os
import pytest
import shutil
import tempfile
from datetime import datetime, timedelta

from insights.core import dr
from insights.core.spec_factory import foreach_collect
from insights.core.context import HostContext
from insights.specs.datasources.eap_reports import eap_report_files


@pytest.fixture
def sample_directory(scope="module"):
    tmpdir = tempfile.mkdtemp()
    os.makedirs(tmpdir + "/var/tmp/insights-runtimes/uploads/")
    for f in ["file_a.json", "file_b.json"]:
        fd = open(tmpdir + "/var/tmp/insights-runtimes/uploads/" + f, "w")
        fd.write('{"name":"example_json", "count_number":30}')
        fd.close()
    yield tmpdir
    shutil.rmtree(tmpdir)


@pytest.fixture
def sample_directory_w_extra_files(scope="module"):
    now = datetime.today()
    tmpdir = tempfile.mkdtemp()
    f_dir = tmpdir + "/var/tmp/insights-runtimes/uploads/"
    os.makedirs(f_dir)
    for f in ['file_%s.json' % str(i) for i in range(12)]:
        fd = open(f_dir + f, "w")
        fd.write('{"name":"example_json", "count_number":30}')
        fd.close()
    # to diff the file time explicitly
    os.system(
        'touch -d "{0}" '.format((now - timedelta(minutes=2)).isoformat()) + f_dir + "file_0.json"
    )
    os.system(
        'touch -d "{0}" '.format((now - timedelta(minutes=1)).isoformat()) + f_dir + "file_1.json"
    )
    yield tmpdir
    shutil.rmtree(tmpdir)


def run_eap_files_test(sample_directory, spec):
    ctx = HostContext()
    ctx.root = sample_directory
    broker = dr.Broker()
    broker[HostContext] = ctx
    broker = dr.run([spec], broker)
    return broker


def test_eap_match_results(sample_directory):
    spec = eap_report_files
    broker = run_eap_files_test(sample_directory, spec)
    file_names = [p.split('/')[-1] for p in broker[spec]]
    assert 'file_a.json' in file_names
    assert 'file_b.json' in file_names


def test_foreach_eap_match_results(sample_directory):
    spec = foreach_collect(eap_report_files, "%s")
    broker = run_eap_files_test(sample_directory, spec)
    file_names = [p.file_name for p in broker[spec]]
    assert 'file_a.json' in file_names
    assert 'file_b.json' in file_names


def test_eap_one_match_paths(sample_directory):
    spec = eap_report_files
    now = datetime.today()
    previous_day = now - timedelta(days=1)
    os.system(
        'touch -d "{0}" '.format(previous_day.isoformat())
        + sample_directory
        + "/var/tmp/insights-runtimes/uploads/file_a.json"
    )
    broker = run_eap_files_test(sample_directory, spec)
    file_names = [p for p in broker[spec]]
    assert len(broker[spec]) == 1
    assert 'file_b.json' in file_names[-1]


def test_eap_zero_match_paths(sample_directory):
    spec = eap_report_files
    now = datetime.today()
    previous_day = now - timedelta(days=1)
    os.system(
        'touch -d "{0}" '.format(previous_day.isoformat())
        + sample_directory
        + "/var/tmp/insights-runtimes/uploads/file_a.json"
    )
    os.system(
        'touch -d "{0}" '.format(previous_day.isoformat())
        + sample_directory
        + "/var/tmp/insights-runtimes/uploads/file_b.json"
    )
    broker = run_eap_files_test(sample_directory, spec)
    assert spec not in broker


def test_eap_reports_limited_on_latest_count(sample_directory_w_extra_files):
    spec = eap_report_files
    ctx = HostContext()
    ctx.root = sample_directory_w_extra_files
    broker = dr.Broker()
    broker[HostContext] = ctx
    broker = dr.run([spec], broker)
    file_names = [p.split('/')[-1] for p in broker[spec]]
    assert len(broker[spec]) == 10
    assert 'file_11.json' in file_names
    assert 'file_2.json' in file_names



# -*- coding: UTF-8 -*-

from insights.client.constants import InsightsConstants
from insights.client.phase.v1 import pre_update
from mock.mock import patch
from pytest import raises, mark


def patch_insights_config(old_function):
    patcher = patch("insights.client.phase.v1.InsightsConfig",
                    **{"return_value.load_all.return_value.auto_config": False,
                       "return_value.load_all.return_value.version": False,
                       "return_value.load_all.return_value.validate": False,
                       "return_value.load_all.return_value.enable_schedule": False,
                       "return_value.load_all.return_value.disable_schedule": False,
                       "return_value.load_all.return_value.analyze_container": False,
                       "return_value.load_all.return_value.test_connection": False,
                       "return_value.load_all.return_value.support": True})
    return patcher(old_function)


def patch_insights_client(old_function):
    patcher = patch("insights.client.phase.v1.InsightsClient")
    return patcher(old_function)


def patch_insights_support(old_function):
    patcher = patch("insights.client.phase.v1.InsightsSupport")
    return patcher(old_function)


@mark.skip(reason="Mock errors in QE Jenkins")
@patch_insights_support
@patch_insights_client
@patch_insights_config
def test_support_called(insights_config, insights_client, insights_support):
    """
    InsightsSupport is constructed with InsightsConfig and collect_support_info is called.
    """
    try:
        pre_update()
    except SystemExit:
        pass

    insights_support.assert_called_once_with(insights_config.return_value.load_all.return_value)
    insights_support.return_value.collect_support_info.assert_called_once_with()


@mark.skip(reason="Mock errors in QE Jenkins")
@patch_insights_support
@patch_insights_client
@patch_insights_config
def test_exit_ok(insights_config, insights_client, insights_support):
    """
    Support collection replaces the normal client run.
    """
    with raises(SystemExit) as exc_info:
        pre_update()

    assert exc_info.value.code == InsightsConstants.sig_kill_ok
